/*! ********************************************
 * MultiTimberSynthBase.cpp
 * Copyright 2019 masahiro takamoto (MotChan)
 * Distributed under the MIT license. 
 * see -> sora/license.txt
 ******************************************** */
#ifndef INCLUDE_MultiTimberSynthBase_h
#include "MultiTimberSynthBase.hpp"
#endif

#include <cmath>

namespace sora {
namespace graybox {
namespace synth {
/***********************************************
	Class	VoicePart
************************************************/
void VoicePart::reset()
{
	programChange = 0;
	chPressure    = 0;
	pitchBend     = 8192;
	
	for (int i = 0; i < 120; ++i) ccValues[i] = 0;
	ccValues[ 1] = 0;
	ccValues[ 7] = 100;
	ccValues[10] = 64;
	ccValues[11] = 127;
	
	rpn         = 0;
	nrpn        = 0;
	paramNumber = 0x100;
	
	bendSens = 2;
	extVal0 = 0;
	extVal1 = 0;
}
void VoicePart::resetAllControlle()
{
	chPressure   = 0;
	pitchBend    = 8192;
	
	ccValues[ 1] = 0;
	ccValues[11] = 127;
	
	rpn         = 0;
	nrpn        = 0;
	paramNumber = 0x100;
}
void VoicePart::resetRPNControlle()
{
	chPressure   = 0;
	pitchBend    = 8192;
	
	ccValues[ 1] = 0;
	ccValues[11] = 127;
	
	rpn         = 0;
	nrpn        = 0;
	paramNumber = 0x100;
}
VoicePart::VoicePart(int32 defaultMode)
{
	this->defaultMode = defaultMode;
	mode = defaultMode;
	reset();
}
/***********************************************
	Class	MultiTimberSynthBase
************************************************/
// Tables -----
void MultiTimberSynthBase::_resetVolumeTables()
{
	if (_masterVolumeTable == nullptr)_masterVolumeTable = new float[128];
	for (int i = 0; i < 128; ++i)_masterVolumeTable[i] = std::pow(2.0f, (i + 1) / 8.0f) / 65536.0f;
}
	// getter, setter ---
void MultiTimberSynthBase::setMasterVolume(float val)
{
	_masterVolume = val;
	for (int i = 0; i < _maxPolyphony; ++i)_voiceUnits[i]->setMasterVolume(_masterVolume);
}
void MultiTimberSynthBase::setSampleRate(float val){ _sampleRate = val; _resetKeyParams(); }
void MultiTimberSynthBase::setReferencePitch(float val){ _referencePitch = val; _resetKeyParams(); }
int32 MultiTimberSynthBase::getPolyphony()
{
	int _cnt = 0;
	for (int i = 0; i < _maxPolyphony; ++i) if (_voiceUnits[i]->state)_cnt++;
	return _cnt;
}
int32 MultiTimberSynthBase::setMaxPolyphony(int32 num)
{
	if (num < 1 || num == _maxPolyphony) return _maxPolyphony;
	if (num >= 1024) num = 1024;
	else for (int v = 1; v < 1024; v <<= 1) if (v >= num){ num = v; break; } 
	_nton = 0;
	_ntonM = num - 1;
	_recreateVU(num);
	_resetKeyParams();
	return _maxPolyphony;
}
// Constructor
MultiTimberSynthBase::MultiTimberSynthBase(int32 maxPoly, float refPitch, float masVol)
{
	_referencePitch = refPitch;
	_masterVolume = masVol;
// Table ---
	_resetVolumeTables();
// RPN, CC ---
	_set_rpnFunc();
	_set_ccFunc();
// Part ---
	if (voiceParts == nullptr)voiceParts = new VoicePart[16];
	for (int i = 0; i < 16; ++i) voiceParts[i].mode = voiceParts[i].defaultMode = 1;
// VoiceUnit ---
	if (maxPoly >= 1024) maxPoly = 1024;
	else for (int v = 1; v < 1024; v <<= 1) if (v >= maxPoly){ maxPoly = v; break; } 
	_nton = 0;
	_ntonM = maxPoly - 1;
	_maxPolyphony = maxPoly;
}
MultiTimberSynthBase::MultiTimberSynthBase(MultiTimberSynthBase && x) noexcept
{
// Table ---
	_masterVolumeTable = x._masterVolumeTable; x._masterVolumeTable = nullptr;
// RPN, CC ---
	_set_rpnFunc();
	_set_ccFunc();
	// Synth Status
	midiInPorts = x.midiInPorts;
	midiOutPorts = x.midiOutPorts;
	receiveSysEx = x.receiveSysEx;
	_manufacID = x._manufacID;
	_deiceID = x._deiceID;
	_modelID = x._modelID;
	// Channel Paramters
	voiceParts = x.voiceParts; x.voiceParts = nullptr;

	_sampleRate = x._sampleRate;
	_referencePitch = x._referencePitch;
	_masterVolume = x._masterVolume;
	// VoiceUnits
	_maxPolyphony = x._maxPolyphony;
	_voiceUnits = x._voiceUnits; x._voiceUnits = nullptr;
	_nton = x._nton;
	_ntonM = x._ntonM;
// MidiCtrl ---
	//_evBuff = x._evBuff;
// RPN,NRPN -----
	_rpnCtrlFlag = x._rpnCtrlFlag;
	_nrpnCtrlFlag = x._nrpnCtrlFlag;
}
MultiTimberSynthBase::~MultiTimberSynthBase()
{
	if (_masterVolumeTable != nullptr)delete[] _masterVolumeTable;
	_masterVolumeTable = nullptr;
	if (voiceParts != nullptr)delete[] voiceParts;
	voiceParts = nullptr;
	if (_voiceUnits != nullptr)delete[] _voiceUnits;
	_voiceUnits = nullptr;
}
// Controls ---
void MultiTimberSynthBase::initialize()
{
	_resetVolumeTables();
	allSoundStop();
	allPartReset(true);
	factoryDataReset();
	_rpnCtrlFlag = true;
	_nrpnCtrlFlag = false;
	receiveSysEx = false;
}
void MultiTimberSynthBase::setDefaultPartMode(int32 part, int32 mode)
{
	if (mode > _maxPMN)mode = _maxPMN;
	voiceParts[part].defaultMode = mode;
}
void MultiTimberSynthBase::setPartMode(int32 part, int32 mode)
{
	if (mode > _maxPMN) mode = _maxPMN;
	if (mode == voiceParts[part].mode) return;
	voiceParts[part].mode = mode;
	allSoundStop(part);
}
void MultiTimberSynthBase::allPartReset(bool modeReset)
{
	for (int i = 0; i < 16; ++i) voiceParts[i].reset();
	if (modeReset)for (int i = 0; i < 16; ++i)voiceParts[i].mode = voiceParts[i].defaultMode;
}
void MultiTimberSynthBase::allSoundStop(int32 part)
{
	if (part < 0) for (int i = 0; i < _maxPolyphony; ++i)_voiceUnits[i]->killProcess();
	else for (int i = 0; i < _maxPolyphony; ++i)if (_voiceUnits[i]->partNumber == part)_voiceUnits[i]->killProcess();
}
void MultiTimberSynthBase::allSoundOff(int32 part)
{
	if (part < 0) for (int i = 0; i < _maxPolyphony; ++i)_voiceUnits[i]->soundOff();
	else for (int i = 0; i < _maxPolyphony; ++i)if (_voiceUnits[i]->partNumber == part)_voiceUnits[i]->soundOff();
}
// MidiCtrl ---
void MultiTimberSynthBase::noteOff(const int32 & ch, const int32 & key)
{
	//_evBuff.type = 0x80;
	_evBuff.channel = ch;
	_evBuff.value0 = key;
	_noteOff(_evBuff);
}
void MultiTimberSynthBase::noteOn(const int32 & ch, const int32 & key, const int32 & velo)
{
	if (voiceParts[ch].mode == 0) return;
	//_evBuff.type = 0x90;
	_evBuff.channel = ch;
	_evBuff.value0 = key;
	_evBuff.value1 = velo;
	if (velo > 0)_noteOn(_evBuff);
	else _noteOff(_evBuff);
}
void MultiTimberSynthBase::keyAfterTouch(const int32 & ch, const int32 & key, const int32 & val)
{
	//_evBuff.type = 0xA0;
	_evBuff.channel = ch;
	_evBuff.value0 = key;
	_evBuff.value1 = val;
	_keyAfterTouch(_evBuff);
}
void MultiTimberSynthBase::controleChange(const int32 & ch, const int32 & cc, const int32 & val)
{
	//_evBuff.type = 0xB0;
	_evBuff.channel = ch;
	//_evBuff.value0 = cc;
	_evBuff.value1 = val;
	(this->*(this->_CC_FUNC[cc]))(_evBuff);
}
void MultiTimberSynthBase::programChange(const int32 & ch, const int32 & num)
{
	//_evBuff.type = 0xC0;
	_evBuff.channel = ch;
	_evBuff.value0 = num;
	_programChange(_evBuff);
}
void MultiTimberSynthBase::channelAfterTouch(const int32 & ch, const int32 & val)
{
	//_evBuff.type = 0xD0;
	_evBuff.channel = ch;
	_evBuff.value0 = val;
	_channelAfterTouch(_evBuff);
}
void MultiTimberSynthBase::pitchBendChange(const int32 & ch, const int32 & lsb, const int32 & msb)
{
	//_evBuff.type = 0xE0;
	_evBuff.channel = ch;
	_evBuff.value0 = lsb;
	_evBuff.value1 = msb;
	_pitchBendChange(_evBuff);
}
void MultiTimberSynthBase::pitchBendChangeW(const int32 & ch, const int32 & val)
{
	//_evBuff.type = 0xE0;
	_evBuff.channel = ch;
	_evBuff.value0 = val & 0x7f;
	_evBuff.value1 = val >> 7 & 0x7f;
	_pitchBendChange(_evBuff);
}
void MultiTimberSynthBase::sysExMessage(const uint8* msg, const int32& dataSize)
{
	if (!receiveSysEx) return;
	//_evBuff.type = 0xF0;
	_evBuff.channel = 0;
	int _size = (dataSize > MidiEvent::DATA_PACKET_SIZE) ? MidiEvent::DATA_PACKET_SIZE : dataSize;
	for (int i = 0; i < _size; ++i)_evBuff.data[i] = msg[i];
	_evBuff.dataSize = _size;
	_readSysExData(_evBuff);
}
// Interface - IfGrayBoxModule -----
void MultiTimberSynthBase::updateSampleRate(float sampleRate)
{
	if (sampleRate == _sampleRate) return;
	setSampleRate(sampleRate);
}
void MultiTimberSynthBase::resetModule()
{
	allSoundStop();
	allPartReset();
}
void MultiTimberSynthBase::initiModule(){ initialize(); }
void MultiTimberSynthBase::connectModule(GrayBox* host){ setSampleRate(host->getSampleRate()); }
void MultiTimberSynthBase::disconnectModule(){ allSoundStop(); }
void MultiTimberSynthBase::updateBypassStatus(bool){ allSoundStop(); }
void MultiTimberSynthBase::startProcess(){ allSoundStop(); }
void MultiTimberSynthBase::stopProcess(){ allSoundStop(); }
void MultiTimberSynthBase::suspendProcess(){ allSoundStop(); }
void MultiTimberSynthBase::resumeProcess(){ allSoundStop(); }
void MultiTimberSynthBase::process(ProcessData* procData)
{
	MidiEvent* _events = procData->midiEventList->next;
	float*& _outL = procData->audioOut[0];
	float*& _outR = procData->audioOut[1];
	// Prepare Process -->
	int _mxp = _maxPolyphony, _startFrame = procData->offsetFrames, _pos = _startFrame;
	VoiceUnitBase** const& _vus = _voiceUnits;
	for (int i = 0; i < _mxp; ++i)_vus[i]->startFrame = _startFrame;
	// Process Events -->
	while (_events != nullptr)
	{
		if ((_events->midiPorts & midiInPorts) == 0){ _events = _events->next; continue; }
		switch (_events->type)
		{
			case 0x80:
				if (_events->offsetFrames > _pos)
				{
					_pos = _events->offsetFrames;
					for (int i = 0; i < _mxp; ++i)
					{
						VoiceUnitBase* const& _vu = _vus[i];
						if (_vu->state)(_vu->*(_vu->process))(_outL, _outR, _pos);
					}
				}
				_noteOff(*_events);
				break;
			case 0x90:
				if (voiceParts[_events->channel].mode == 0) break;
				if (_events->offsetFrames > _pos)
				{
					_pos = _events->offsetFrames;
					for (int i = 0; i < _mxp; ++i)
					{
						VoiceUnitBase* const& _vu = _vus[i];
						if (_vu->state)(_vu->*(_vu->process))(_outL, _outR, _pos);
					}
				}
				if (_events->value1 > 0)_noteOn(*_events);
				else _noteOff(*_events);
				break;
			case 0xA0:
				if (_events->offsetFrames > _pos)
				{
					_pos = _events->offsetFrames;
					for (int i = 0; i < _mxp; ++i)
					{
						VoiceUnitBase* const& _vu = _vus[i];
						if (_vu->state)(_vu->*(_vu->process))(_outL, _outR, _pos);
					}
				}
				_keyAfterTouch(*_events);
				break;
			case 0xB0:
				if (_events->offsetFrames > _pos)
				{
					_pos = _events->offsetFrames;
					for (int i = 0; i < _mxp; ++i)
					{
						VoiceUnitBase* const& _vu = _vus[i];
						if (_vu->state)(_vu->*(_vu->process))(_outL, _outR, _pos);
					}
				}
				(this->*(this->_CC_FUNC[_events->value0]))(*_events);
				break;
			case 0xC0:
				if (_events->offsetFrames > _pos)
				{
					_pos = _events->offsetFrames;
					for (int i = 0; i < _mxp; ++i)
					{
						VoiceUnitBase* const& _vu = _vus[i];
						if (_vu->state)(_vu->*(_vu->process))(_outL, _outR, _pos);
					}
				}
				voiceParts[_events->channel].programChange = _events->value0;
				break;
			case 0xD0:
				if (_events->offsetFrames > _pos)
				{
					_pos = _events->offsetFrames;
					for (int i = 0; i < _mxp; ++i)
					{
						VoiceUnitBase* const& _vu = _vus[i];
						if (_vu->state)(_vu->*(_vu->process))(_outL, _outR, _pos);
					}
				}
				_channelAfterTouch(*_events);
				break;
			case 0xE0:
				if (_events->offsetFrames > _pos)
				{
					_pos = _events->offsetFrames;
					for (int i = 0; i < _mxp; ++i)
					{
						VoiceUnitBase* const& _vu = _vus[i];
						if (_vu->state)(_vu->*(_vu->process))(_outL, _outR, _pos);
					}
				}
				_pitchBendChange(*_events);
				break;
			case 0xF0:
				if (!receiveSysEx) break;
				if (_events->offsetFrames > _pos)
				{
					_pos = _events->offsetFrames;
					for (int i = 0; i < _mxp; ++i)
					{
						VoiceUnitBase* const& _vu = _vus[i];
						if (_vu->state)(_vu->*(_vu->process))(_outL, _outR, _pos);
					}
				}
				_readSysExData(*_events);
				break;
		}
		_events = _events->next;
	}
	// End Process -->
	int _endFrame = _startFrame + procData->sampleFrames;
	for (int i = 0; i < _mxp; ++i)
	{
		VoiceUnitBase* const& _vu = _vus[i];
		if (_vu->state)(_vu->*(_vu->process))(_outL, _outR, _endFrame);
	}
}
// MidiEvent ---
void MultiTimberSynthBase::_noteOff(const MidiEvent &ev)
{
	int _key = ev.value0, _ch = ev.channel;
	for (int i = 0; i < _maxPolyphony; ++i)
	{
		VoiceUnitBase* const& _vu = _voiceUnits[i];
		// KeyOn -->
		if (_vu->state & 0x010) if (_vu->partNumber == _ch && _vu->noteNumber == _key)_vu->keyOff();
	}
}
void MultiTimberSynthBase::_noteOn(const MidiEvent &ev)
{
	for (int i = 0; i < _maxPolyphony; ++i)
	{
		VoiceUnitBase* const& _vu = _voiceUnits[i];
		if (_vu->state == 0)
		{
			_vu->startFrame = ev.offsetFrames;
			_vu->keyOn(ev.channel, ev.value0, ev.value1);
			return;
		}
	}
	int _len = _nton + _maxPolyphony;
	for (int i = _nton; i < _len; ++i)
	{
		VoiceUnitBase* const& _vu = _voiceUnits[i & _ntonM];
		// (KeyOff || SoundOff) -->
		if (_vu->state & 0x60)
		{
			_nton = _nton + 1 & _ntonM;
			_vu->startFrame = ev.offsetFrames;
			_vu->keyOn(ev.channel, ev.value0, ev.value1);
			return;
		}
	}
	for (int i = _nton; i < _len; ++i)
	{
		VoiceUnitBase* const& _vu = _voiceUnits[i & _ntonM];
		// KeyOn -->
		if (_vu->state & 0x10)
		{
			_nton = _nton + 1 & _ntonM;
			_vu->startFrame = ev.offsetFrames;
			_vu->keyOn(ev.channel, ev.value0, ev.value1);
			return;
		}
	}
}
void MultiTimberSynthBase::_programChange(const MidiEvent &ev)
{
	int _ch = ev.channel;
	voiceParts[_ch].programChange = ev.value0;
	if (voiceParts[_ch].mode == 0) return;
	for (int i = 0; i < _maxPolyphony; ++i) if (_voiceUnits[i]->partNumber == _ch)_voiceUnits[i]->soundOff();
}
void MultiTimberSynthBase::_channelAfterTouch(const MidiEvent &ev)
{
	voiceParts[ev.channel].chPressure = ev.value0;
}
void MultiTimberSynthBase::_pitchBendChange(const MidiEvent &ev)
{
	int _ch = ev.channel;
	VoicePart& _pt = voiceParts[_ch];
	_pt.pitchBend = ev.value1 << 7 | ev.value0;
	if (_pt.mode == 0) return;
	for (int i = 0; i < _maxPolyphony; ++i)
	{
		VoiceUnitBase* const& _vu = _voiceUnits[i];
		if (_vu->partNumber == _ch)_vu->updPartPitchBend(_pt.pitchBend);
	}
}
void MultiTimberSynthBase::_set_ccFunc()
{
	for (int i =   0; i < 120; ++i)_CC_FUNC[i] = &MultiTimberSynthBase::_ccFunc_channelVoiceUB;
	for (int i = 120; i < 128; ++i)_CC_FUNC[i] = &MultiTimberSynthBase::_ccFunc_channelModeUB;
	_CC_FUNC[0x01] = &MultiTimberSynthBase::_ccFunc_01;
	_CC_FUNC[0x06] = &MultiTimberSynthBase::_ccFunc_06;
	_CC_FUNC[0x07] = &MultiTimberSynthBase::_ccFunc_07;
	_CC_FUNC[0x0A] = &MultiTimberSynthBase::_ccFunc_0A;
	_CC_FUNC[0x0B] = &MultiTimberSynthBase::_ccFunc_0B;
	_CC_FUNC[0x26] = &MultiTimberSynthBase::_ccFunc_26;
	_CC_FUNC[0x60] = &MultiTimberSynthBase::_ccFunc_60;
	_CC_FUNC[0x61] = &MultiTimberSynthBase::_ccFunc_61;
	_CC_FUNC[0x62] = &MultiTimberSynthBase::_ccFunc_62;
	_CC_FUNC[0x63] = &MultiTimberSynthBase::_ccFunc_63;
	_CC_FUNC[0x64] = &MultiTimberSynthBase::_ccFunc_64;
	_CC_FUNC[0x65] = &MultiTimberSynthBase::_ccFunc_65;
	_CC_FUNC[0x78] = &MultiTimberSynthBase::_ccFunc_78;
	_CC_FUNC[0x79] = &MultiTimberSynthBase::_ccFunc_79;
	_CC_FUNC[0x7B] = &MultiTimberSynthBase::_ccFunc_7B;
}
// MidiEvent - CC -----
void MultiTimberSynthBase::_ccFunc_channelVoiceUB(const MidiEvent &ev)
{
	voiceParts[ev.channel].ccValues[ev.value0] = ev.value1;
}
void MultiTimberSynthBase::_ccFunc_01(const MidiEvent &ev)
{
	// Modulation --->
	int _ch = ev.channel;
	VoicePart& _pt = voiceParts[_ch];
	_pt.ccValues[1] = ev.value1;
	if (_pt.mode == 0) return;
	for (int i = 0; i < _maxPolyphony; ++i)
	{
		VoiceUnitBase* const& _vu = _voiceUnits[i];
		if (_vu->partNumber == _ch)_vu->updPartModulation(ev.value1);
	}
}
void MultiTimberSynthBase::_ccFunc_06(const MidiEvent &ev)
{
	// Data Entry MSB --->
	voiceParts[ev.channel].ccValues[6] = ev.value1;
	(this->*(this->_RPN_FUNC[voiceParts[ev.channel].paramNumber]))(ev.channel, ev.value1 << 7, 0x7f, 0);
}
void MultiTimberSynthBase::_ccFunc_07(const MidiEvent &ev)
{
	// Volume --->
	int _ch = ev.channel;
	VoicePart& _pt = voiceParts[_ch];
	_pt.ccValues[7] = ev.value1;
	if (_pt.mode == 0) return;
	for (int i = 0; i < _maxPolyphony; ++i)
	{
		VoiceUnitBase* const& _vu = _voiceUnits[i];
		if (_vu->partNumber == _ch)_vu->updPartVolume(ev.value1);
	}
}
void MultiTimberSynthBase::_ccFunc_0A(const MidiEvent &ev)
{
	// Pan --->
	int _ch = ev.channel;
	VoicePart& _pt = voiceParts[_ch];
	_pt.ccValues[10] = ev.value1;
	if (_pt.mode == 0) return;
	for (int i = 0; i < _maxPolyphony; ++i)
	{
		VoiceUnitBase* const& _vu = _voiceUnits[i];
		if (_vu->partNumber == _ch)_vu->updPartPan(ev.value1);
	}
}
void MultiTimberSynthBase::_ccFunc_0B(const MidiEvent &ev)
{
	// Expression --->
	int _ch = ev.channel;
	VoicePart& _pt = voiceParts[_ch];
	_pt.ccValues[11] = ev.value1;
	if (_pt.mode == 0) return;
	for (int i = 0; i < _maxPolyphony; ++i)
	{
		VoiceUnitBase* const& _vu = _voiceUnits[i];
		if (_vu->partNumber == _ch)_vu->updPartExpression(ev.value1);
	}
}
void MultiTimberSynthBase::_ccFunc_26(const MidiEvent &ev)
{
	// Data Entry LSB --->
	voiceParts[ev.channel].ccValues[38] = ev.value1;
	(this->*(this->_RPN_FUNC[voiceParts[ev.channel].paramNumber]))(ev.channel, ev.value1, 0x3f80, 0);
}
void MultiTimberSynthBase::_ccFunc_60(const MidiEvent &ev)
{
	// Data Inc --->
	(this->*(this->_RPN_FUNC[voiceParts[ev.channel].paramNumber]))(ev.channel, 0, 0x3fff, 1);
}
void MultiTimberSynthBase::_ccFunc_61(const MidiEvent &ev)
{
	// Data Dec --->
	(this->*(this->_RPN_FUNC[voiceParts[ev.channel].paramNumber]))(ev.channel, 0, 0x3fff, -1);
}
void MultiTimberSynthBase::_ccFunc_62(const MidiEvent &ev)
{
	// NRPN LSB --->
	if (!_nrpnCtrlFlag) return;
	VoicePart& _pt = voiceParts[ev.channel];
	_pt.ccValues[98] = ev.value1;
	_pt.nrpn = 0x4000 | _pt.nrpn & 0xff80 | ev.value1;
	if (_pt.nrpn >= 0xc000)  // Received MSB,LSB
	{
		_pt.paramNumber = _pt.nrpn & 0x3fff;
		if (_pt.paramNumber <= 0x7f)_pt.paramNumber |= 0x80;
		else _pt.paramNumber = 0x100;
	}
}
void MultiTimberSynthBase::_ccFunc_63(const MidiEvent &ev)
{
	// NRPN MSB --->
	if (!_nrpnCtrlFlag) return;
	VoicePart& _pt = voiceParts[ev.channel];
	_pt.ccValues[99] = ev.value1;
	_pt.nrpn = 0x8000 | _pt.nrpn & 0xc07f | ev.value1 << 7;
	if (_pt.nrpn >= 0xc000)  // Received MSB,LSB
	{
		_pt.paramNumber = _pt.nrpn & 0x3fff;
		if (_pt.paramNumber <= 0x7f)_pt.paramNumber |= 0x80;
		else _pt.paramNumber = 0x100;
	}
}
void MultiTimberSynthBase::_ccFunc_64(const MidiEvent &ev)
{
	// RPN LSB --->
	if (!_rpnCtrlFlag) return;
	VoicePart& _pt = voiceParts[ev.channel];
	_pt.ccValues[100] = ev.value1;
	_pt.rpn = 0x4000 | _pt.rpn & 0xff80 | ev.value1;
	if (_pt.rpn >= 0xc000)  // Received MSB,LSB
	{
		_pt.paramNumber = _pt.rpn & 0x3fff;
		if (_pt.paramNumber > 0x7f)_pt.paramNumber = 0x100;
	}
}
void MultiTimberSynthBase::_ccFunc_65(const MidiEvent &ev)
{
	// RPN MSB --->
	if (!_rpnCtrlFlag) return;
	VoicePart& _pt = voiceParts[ev.channel];
	_pt.ccValues[101] = ev.value1;
	_pt.rpn = 0x8000 | _pt.rpn & 0xc07f | ev.value1 << 7;
	if (_pt.rpn >= 0xc000)  // Received MSB,LSB
	{
		_pt.paramNumber = _pt.rpn & 0x3fff;
		if (_pt.paramNumber > 0x7f)_pt.paramNumber = 0x100;
	}
}
void MultiTimberSynthBase::_ccFunc_78(const MidiEvent &ev)
{
	// All Sound Off --->
	int _ch = ev.channel;
	if (voiceParts[ev.channel].mode == 0) return;
	for (int i = 0; i < _maxPolyphony; ++i)
	{
		VoiceUnitBase* const& _vu = _voiceUnits[i];
		if ( _vu->partNumber == _ch)_vu->soundOff();
	}
}
void MultiTimberSynthBase::_ccFunc_79(const MidiEvent &ev)
{
	// Reset All Controller --->
	int _ch = ev.channel;
	VoicePart& _pt = voiceParts[_ch];
	_pt.resetAllControlle();
	if (_pt.mode == 0) return;
	for (int i = 0; i < _maxPolyphony; ++i)
	{
		VoiceUnitBase* const& _vu = _voiceUnits[i];
		if (_vu->partNumber == _ch)
		{
			_vu->updPartPitchBend(_pt.pitchBend);
			_vu->updPartModulation(_pt.ccValues[1]);
			_vu->updPartExpression(_pt.ccValues[11]);
		}
	}
}
void MultiTimberSynthBase::_ccFunc_7B(const MidiEvent &ev)
{
	// All Note Off --->
	int _ch = ev.channel;
	if (voiceParts[ev.channel].mode == 0) return;
	for (int i = 0; i < _maxPolyphony; ++i)
	{
		VoiceUnitBase* const& _vu = _voiceUnits[i];
		// KeyOn -->
		if (_vu->state & 0x010 && _vu->partNumber == _ch)_vu->keyOff();
	}
}
// RPN,NRPN -----
void MultiTimberSynthBase::setRPNCtrlFlag(bool flag)
{
	_rpnCtrlFlag = flag;
	for (int i = 0; i < 16; ++i) voiceParts[i].resetRPNControlle();
}
void MultiTimberSynthBase::setNRPNCtrlFlag(bool flag)
{
	_nrpnCtrlFlag = flag;
	for (int i = 0; i < 16; ++i) voiceParts[i].resetRPNControlle();
}
void MultiTimberSynthBase::_set_rpnFunc()
{
	for (int i = 0; i < 256; ++i)_RPN_FUNC[i] = &MultiTimberSynthBase::_rpnFunc_UB;
	_RPN_FUNC[0x100] = &MultiTimberSynthBase::_rpnFunc_Null;
	// rpn
	_RPN_FUNC[0] = &MultiTimberSynthBase::_rpnFunc_00;
	_RPN_FUNC[1] = &MultiTimberSynthBase::_rpnFunc_01;
	_RPN_FUNC[2] = &MultiTimberSynthBase::_rpnFunc_02;
	// nrpn
	_RPN_FUNC[0x80] = &MultiTimberSynthBase::_nrpnFunc_00;
	_RPN_FUNC[0x80 | 0x20] = &MultiTimberSynthBase::_nrpnFunc_20;
	_RPN_FUNC[0x80 | 0x21] = &MultiTimberSynthBase::_nrpnFunc_21;
}
void MultiTimberSynthBase::_rpnFunc_00(const int32 &ch, const int32 &param, const int32 &mask, const int32 &inc)
{
	if (mask == 0x3f80) return;  // DataEntry LSB
	// DataEntry MSB, DataInc, DataDec ->
	int _val = (voiceParts[ch].bendSens << 7) & mask | param;
	_val = (_val >> 7) + inc;
	if		(_val <   0) _val = 0; 
	else if (_val > 127) _val = 127; 
	voiceParts[ch].bendSens = _val;
	if (voiceParts[ch].mode == 0) return;
	for (int i = 0; i < _maxPolyphony; ++i)
	{
		VoiceUnitBase* const& _vu = _voiceUnits[i];
		if (_vu->partNumber == ch)_vu->updPartBendSens(_val);
	}
}
void MultiTimberSynthBase::_nrpnFunc_00(const int32 &, const int32 &param, const int32 &mask, const int32 &)
{
	if (mask != 0x7f) return;
	// DataEntry MSB ->
	setMasterVolume(_masterVolumeTable[param >> 7]);
}
void MultiTimberSynthBase::_nrpnFunc_20(const int32 & ch, const int32 &param, const int32 &mask, const int32 &)
{
	if (mask != 0x7f) return;
	// DataEntry MSB ->
	setDefaultPartMode(ch, param >> 7);
}
void MultiTimberSynthBase::_nrpnFunc_21(const int32 & ch, const int32 &param, const int32 &mask, const int32 &)
{
	if (mask != 0x7f) return;
	// DataEntry MSB ->
	setPartMode(ch, param >> 7);
}
/***********************************************
	Class	MultiTimberSynthBase :: VoiceUnitBase
************************************************/
void MultiTimberSynthBase::VoiceUnitBase::setMasterVolume(const float &val)
{
	_masterVolume = val;
	if (state == 0) return;
	_vVol = _vVel * _masterVolume;
	_vOut = _vPtVol * _vPtExp * _vVol;
	_vOutL = _vOut * _vPanL;
	_vOutR = _vOut * _vPanR;
}
void MultiTimberSynthBase::VoiceUnitBase::updPartVolume(const int32 &val)
{
	_vPtVol = _mainVolTable[val];
	_vOut = _vPtVol * _vPtExp * _vVol;
	_vOutL = _vOut * _vPanL;
	_vOutR = _vOut * _vPanR;
}
void MultiTimberSynthBase::VoiceUnitBase::updPartExpression(const int32 &val)
{
	_vPtExp = _exprVolTable[val];
	_vOut = _vPtVol * _vPtExp * _vVol;
	_vOutL = _vOut * _vPanL;
	_vOutR = _vOut * _vPanR;
}
void MultiTimberSynthBase::VoiceUnitBase::updPartPan(const int32 &val)
{
	_vPanL = _panTable[      _prPan + val];
	_vPanR = _panTable[256 - _prPan - val];
	_vOutL = _vOut * _vPanL;
	_vOutR = _vOut * _vPanR;
}

}}}
