/*! ********************************************
 * SampleData.cpp
 * Copyright 2019 masahiro takamoto (MotChan)
 * Distributed under the MIT license. 
 * see -> sora/license.txt
 ******************************************** */
#ifndef INCLUDE_SampleData_h
#include "SampleData.hpp"
#endif

#include <cmath>

namespace sora {
namespace graybox {
namespace synth {
/***********************************************
	Class	SamplerPreset
************************************************/
void SamplerPreset::copy(const SamplerPreset& preset)
{
	for (int i = 0; i < 16; ++i)toneName[i] = preset.toneName[i];
	smplNum = preset.smplNum;
	pan = preset.pan;
	octv = preset.octv;
	tune = preset.tune;
	tl = preset.tl;
	ar = preset.ar;
	dr = preset.dr;
	sl = preset.sl;
	rr = preset.rr;
}
void SamplerPreset::reset()
{
	for (int i = 0; i < 16; ++i)toneName[i] = ' ';
	smplNum = 0;
	pan = 64;
	octv = 8;
	tune = 8192;
	tl = 0;
	ar = 127;
	dr = 0;
	sl = 127;
	rr = 120;
}
/***********************************************
	Class	SampleData
************************************************/
int32 SampleData::_log2VTindex(double val)
{
	if      (val >  _BOADER) return  int(std::log2( val) * _RES + 0.5) << 1;
	else if (val < -_BOADER) return (int(std::log2(-val) * _RES + 0.5) << 1) + 1;  // +1 = Nega
	else return _BOTTOM;
}

void SampleData::setSampleRate(float sampleRate)
{
	_sampleRate = sampleRate;
	pgSample = (sampleRate <= 1.0f)? 0 : int(std::log2f(sampleRate) * _POW2T_OCTV_RESf + 0.5f);
	pgKey = pgSample - pgPitch;
}
// getter, setter
void  SampleData::setRecPitch(float pitch)
{
	_recPitch = pitch;
	pgPitch = (pitch <= 1.0f)? 0 : int(std::log2f(pitch) * _POW2T_OCTV_RESf + 0.5f);
	pgKey = pgSample - pgPitch;
}
// Method
void SampleData::clear()
{
	if (_created)delete[] sampleData;
	_created = false;
	_sampleRate = 0.0f;
	_recPitch = 0.0f;
	sampleFrames = 0;
	channelCount = 0;
	sampleData = nullptr;
	pgSample = 0;
	pgPitch = 0;
	pgKey = 0;
	plbkS = 0;
	plbkE = 0;
	loopE = 0;
	loopF = 0;
	loopFk = 0;
}
void SampleData::copy(const SampleData& from, bool deepCopy)
{
	clear();
	sampleFrames = from.sampleFrames;
	channelCount = from.channelCount;
	_sampleRate = from._sampleRate;
	_recPitch = from._recPitch;
	// Synth Param
	pgSample = from.pgSample;
	pgPitch = from.pgPitch;
	pgKey = from.pgKey;
	plbkS = from.plbkS;
	plbkE = from.plbkE;
	loopE = from.loopE;
	loopF = from.loopF;
	loopFk = from.loopFk;
	// wave
	int32* _source = from.sampleData;
	if (deepCopy && _source != nullptr)
	{
		_created = true;
		const int _len = from.sampleFrames * from.channelCount;
		sampleData = new int32[_len];
		for (int i = 0; i < _len; ++i)sampleData[i] = _source[i];
	}
	else sampleData = _source;
}
void SampleData::setPlaybackRange(int32 start, int32 end)
{
	if (start >= end || end > sampleFrames) return;
	plbkS = start;
	plbkE = end;
	loopE = plbkE;
	loopF = 0;
	loopFk = 0;
}
void SampleData::resetLoopPoints(int32 start, int32 end)
{
	// loop
	if (start >= plbkS && start < end && end <= plbkE)
	{
		loopE = end;
		loopF = end - start;
		loopFk = (end == plbkE)? loopF : 0;
	}
	// no loop
	else
	{
		loopE = plbkE;
		loopF = 0;
		loopFk = 0;
	}
}
void SampleData::loadPCMSamples(float* wave, int32 samples, int32 ch)
{
	clear();
	_created = true;
	// ---
	const int _len = samples * ch;
	int32* _wave = new int32[_len];
	for (int i = 0; i < _len; ++i)_wave[i] = _log2VTindex(wave[i]);
	sampleData = _wave;
	channelCount = ch;
	sampleFrames = samples;
	// pitch
	_sampleRate = 44100.0f;
	setSampleRate(_sampleRate);
	// playback
	plbkE = sampleFrames;
	loopE = sampleFrames;
}
void SampleData::loadPCMData(uint8* dat, int32 sampleFrames, int32 ch, int32 bitRate, bool sign, bool bigEndian)
{
	clear();
	// decode -->
	const int32 _len = sampleFrames * ch;
	int32* _wave;
	double _sl;
	int _p0, _p1, _p2, _p3;
	switch (bitRate)
	{
		default: clear(); return;
		
		case  8:
			_wave = new int32[_len];
			_sl = 1.0 / 128.0;
			if (sign) for (int i = 0; i < _len; ++i) _wave[i] = _log2VTindex(int8(*(dat++)) * _sl);
			else      for (int i = 0; i < _len; ++i) _wave[i] = _log2VTindex(int8(*(dat++) - 128) * _sl);
			break;
			
		case 16:
			_wave = new int32[_len];
			_sl = 1.0 / 32768.0;
			if (bigEndian){ _p0 = 0; _p1 = 1; } else { _p0 = 1; _p1 = 0; }
			if (sign)
			{
				for (int i = 0; i < _len; ++i, dat += 2)
				{
					int16 _v = int16(dat[_p0] << 8 | dat[_p1]);
					_wave[i] = _log2VTindex(_v * _sl);
				}
			}
			else
			{
				for (int i = 0; i < _len; ++i, dat+=2)
				{
					double _v = uint16(dat[_p0] << 8 | dat[_p1]) - 32768.0;
					_wave[i] = _log2VTindex(_v * _sl);
				}
			}
			break;
			
		case 24:
			_wave = new int32[_len];
			_sl = 1.0 / 8388608.0;
			if (bigEndian){ _p0 = 0; _p1 = 1; _p2 = 2; }
			else          { _p0 = 2; _p1 = 1; _p2 = 0; }
			if (sign)
			{
				for (int i = 0; i < _len; ++i, dat += 3)
				{
					int32 _v = int32(dat[_p0] << 24 | dat[_p1] << 16 | dat[_p2] << 8) / 256;// >> 8;
					_wave[i] = _log2VTindex(_v * _sl);
				}
			}
			else
			{
				for (int i = 0; i < _len; ++i, dat+=3)
				{
					double _v = int32(dat[_p0] << 16 | dat[_p1] << 8 | dat[_p2]) - 8388608.0;
					_wave[i] = _log2VTindex(_v * _sl);
				}
			}
			break;
			
		case 32:
			_wave = new int32[_len];
			_sl = 1.0 / 2147483648.0;
			if (bigEndian){ _p0 = 0; _p1 = 1; _p2 = 2; _p3 = 3; }
			else          { _p0 = 3; _p1 = 2; _p2 = 1; _p3 = 0; }
			if (sign)
			{
				for (int i = 0; i < _len; ++i, dat+=4)
				{
					int32 _v = int32(dat[_p0] << 24 | dat[_p1] << 16 | dat[_p2] << 8 | dat[_p3]);
					_wave[i] = _log2VTindex(_v * _sl);
				}
			}
			else
			{
				for (int i = 0; i < _len; ++i, dat+=4)
				{
					double _v = uint32(dat[_p0] << 24 | dat[_p1] << 16 | dat[_p2] << 8 | dat[_p3]) - 2147483648.0;
					_wave[i] = _log2VTindex(_v * _sl);
				}
			}
			break;
	}
	sampleData = _wave;
	channelCount = ch;
	this->sampleFrames = sampleFrames;
	// pitch
	_sampleRate = 44100.0f;
	setSampleRate(_sampleRate);
	// playback
	plbkE = sampleFrames;
	loopE = sampleFrames;
	//
	_created = true;
}
// Copy Constructor
SampleData::SampleData(const SampleData& x)
{
	copy(x, x._created);
}
SampleData& SampleData::operator = (const SampleData& x)
{
	if (this == &x)return *this;
	copy(x, x._created);
	return *this;
}
// Move Constructor
SampleData::SampleData(SampleData&& x)noexcept
{
	_created = x._created; x._created = false;
	sampleData = x.sampleData; x.sampleData = nullptr;
	sampleFrames = x.sampleFrames;
	channelCount = x.channelCount;
	_sampleRate = x._sampleRate;
	_recPitch = x._recPitch;
	// Synth Param
	pgSample = x.pgSample;
	pgPitch = x.pgPitch;
	pgKey = x.pgKey;
	plbkS = x.plbkS;
	plbkE = x.plbkE;
	loopE = x.loopE;
	loopF = x.loopF;
	loopFk = x.loopFk;
}
SampleData& SampleData::operator = (SampleData&& x)noexcept
{
	if (this == &x)return *this;
	clear();
	_created = x._created; x._created = false;
	sampleData = x.sampleData; x.sampleData = nullptr;
	sampleFrames = x.sampleFrames;
	channelCount = x.channelCount;
	_sampleRate = x._sampleRate;
	_recPitch = x._recPitch;
	// Synth Param
	pgSample = x.pgSample;
	pgPitch = x.pgPitch;
	pgKey = x.pgKey;
	plbkS = x.plbkS;
	plbkE = x.plbkE;
	loopE = x.loopE;
	loopF = x.loopF;
	loopFk = x.loopFk;
	return *this;
}
// Destructor
SampleData::~SampleData()
{
	if (_created)delete[] sampleData;
}
}}}