/*! ********************************************
 * SynthesizerFM4.cpp
 * Copyright 2019 masahiro takamoto (MotChan)
 * Distributed under the MIT license. 
 * see -> sora/license.txt
 ******************************************** */
#ifndef INCLUDE_SynthesizerFM4_h
#include "SynthesizerFM4.hpp"
#endif

#include <utility>
#include <cmath>

namespace sora {
namespace graybox {
namespace synth {
/***********************************************
	Class	TonePresetFM
************************************************/
void TonePresetFM4::copy(const TonePresetFM4& preset)
{
	for (int i = 0; i < 16; ++i)toneName[i] = preset.toneName[i];
	key = preset.key;
	pan = preset.pan;
	lfoWave = preset.lfoWave;
	lfoFreq = preset.lfoFreq;		
	fl = preset.fl;
	strc = preset.strc;
	for (int i = 0; i < 4; ++i)
	{
		tlVelCtrl[i] = preset.tlVelCtrl[i];
		mul[i] = preset.mul[i];
		tune[i] = preset.tune[i];
		lms[i] = preset.lms[i];
		wave[i] = preset.wave[i];
		tl[i] = preset.tl[i];
		ar[i] = preset.ar[i];
		dr[i] = preset.dr[i];
		sl[i] = preset.sl[i];
		rr[i] = preset.rr[i];
		kf[i] = preset.kf[i];
		kfStart[i] = preset.kfStart[i];
	}
}
void TonePresetFM4::reset()
{
	for (int i = 0; i < 16; ++i)toneName[i] = ' ';
	key = 64;
	pan = 64;
	lfoWave = 0;
	lfoFreq = 2048;
	fl = 127;
	strc = 0;
	for (int i = 0; i < 4; ++i)
	{
		tlVelCtrl[i] = 0;
		mul[i] = 1;
		tune[i] = 8192;
		lms[i] = 0;
		wave[i] = 0;
		tl[i] = 0;
		ar[i] = 124;
		dr[i] = 100;
		sl[i] = 0;
		rr[i] = 120;
		kf[i] = 0;
		kfStart[i] = 48;
	}
}
/***********************************************
	Class	SynthesizerFM4
************************************************/
void SynthesizerFM4::_recreateVU(int32 maxPoly)
{
	// Recreate Array
	if (_voiceUnits != nullptr)delete[] _voiceUnits;
	_voiceUnits = new VoiceUnitBase*[maxPoly];
	if (_voiceUnitExt != nullptr)delete[] _voiceUnitExt;
	_voiceUnitExt = new VoiceUnitFM4[maxPoly];
	// Reset VoiceUnit
	for (int i = 0; i < maxPoly; ++i)
	{
		VoiceUnitFM4& _vu = _voiceUnitExt[i];
		_voiceUnits[i] = &_voiceUnitExt[i];
		_vu.setup(_synthData);
		_vu.killProcess();
		_vu.VOICE_PARTS = voiceParts;
		_vu.WAVE_TABLES = waveTables;
		_vu.SAMPLE_DATAS = sampleDatas;
		_vu.TONE_PRESETS = tonePresets;
		_vu.SAMPLER_PRESETS = samplerPresets;
		_vu.setMasterVolume(_masterVolume);
	}
	// set Property
	_maxPolyphony = maxPoly;
}
void SynthesizerFM4::_resetKeyParams()
{
	const double _sampleRateD = double(_sampleRate);
	// LFO
	double _lfoWaveSize = SynthData::LFO_WAVE_TABLE_SIZE / 2.0;
		// Max Frame Rate - LFO Phase Step
	double _lfoMaxFrameRate = _sampleRateD / SynthData::LFO_MAX_FREQ / _lfoWaveSize;
	if (_lfoMaxFrameRate < 1.0)_lfoMaxFrameRate = 1.0;  // Max 1frame per 1Phase
	int _lfoTimerCoef = int(SynthData::LFO_TIMER_INIT_VALUE / 16383.0 / _lfoMaxFrameRate + 0.5);
		
	// EG
	double _rate0Frames = SynthData::EG_REFERENCE_TIME * SynthData::LOG2VT_DYN_BITS * _sampleRateD;  // to -96dB Frames
	int _egPow2tIdx = SynthData::toPow2tIndexValue(SynthData::EG_MAX_CNT / _rate0Frames);
	double _sndOffFrames = SynthData::EG_SOUND_OFF_TIME / SynthData::LOG2VT_DYN_BITS * _sampleRateD;  // to -96dB Frames
	int _egSndOff = int(SynthData::EG_MAX_CNT / _sndOffFrames + 0.5);
	// PG
	int _smpl = SynthData::toPow2tIndexValue(_sampleRateD);
	int _pgSamplePow2tIdx = SynthData::toPow2tIndexValue(double(1 << PG_PHASE_FP)) - _smpl;
	// PG WaveTable
	int _pgKeyPow2tIdx = SynthData::toPow2tIndexValue(double(WAVE_TABLE_SIZE << PG_PHASE_FP)) - _smpl;
	_pgKeyPow2tIdx += SynthData::toPow2tIndexValue(_referencePitch * std::pow(2.0, -69.0 / 12.0)); // note 0 Pitch
	for (int i = 0; i < _maxPolyphony; ++i)
	{
		VoiceUnitFM4& _vu = _voiceUnitExt[i];
		_vu.killProcess();
		_vu.pgSamplePow2tIdx = _pgSamplePow2tIdx;
		_vu.pgKeyPow2tIdx = _pgKeyPow2tIdx;
		_vu.egPow2tIdx = _egPow2tIdx;
		_vu.egPhaseStepSndOff = _egSndOff;
		_vu.lfoTimerCoef = _lfoTimerCoef;
	}
}
void SynthesizerFM4::getWaveForm(int32 id, float* buffer, int32 bufferSize)
{
	if (bufferSize > WAVE_TABLE_SIZE)bufferSize = WAVE_TABLE_SIZE;
	// ---
	const float* _log2vt = _synthData->getLOG2V_TABLE();
	const int32* _table = &waveTables[id << WAVE_TABLE_SIZE_BITS];
	for (int i = 0; i < bufferSize; ++i)buffer[i] = _log2vt[_table[i]];
}
void SynthesizerFM4::setWaveForm(int32 id, const float* buffer, int32 bufferSize)
{
	if (bufferSize > WAVE_TABLE_SIZE)bufferSize = WAVE_TABLE_SIZE;
	const float _RES = float(-SynthData::LOG2VT_DYN_RES),
		_BOADER = 1.0f / (1 << SynthData::LOG2VT_DYN_BITS);
	const int _BOTTOM = SynthData::LOG2VT_SIZE;
	int32* _table = &waveTables[id << WAVE_TABLE_SIZE_BITS];
	for (int i = 0; i < bufferSize; ++i)
	{
		float v = buffer[i];
		if      (v >  _BOADER) _table[i] =  int(std::log2f( v) * _RES + 0.5f) << 1;
		else if (v < -_BOADER) _table[i] = (int(std::log2f(-v) * _RES + 0.5f) << 1) + 1;
		else _table[i] = _BOTTOM;
	}
}
void SynthesizerFM4::setSampleData(int32 id, SampleData&& sampleData)
{
	allSoundStop();
	sampleDatas[id] = std::move(sampleData);
}
void SynthesizerFM4::setSampleData(int32 id, const SampleData& sampleData, bool deepCopy)
{
	allSoundStop();
	sampleDatas[id].copy(sampleData, deepCopy);
}
void SynthesizerFM4::loadPCMSamples(int32 id, float* wave, int32 samples, int32 ch)
{
	allSoundStop();
	sampleDatas[id].loadPCMSamples(wave, samples, ch);
}
void SynthesizerFM4::loadPCMData(int32 id, uint8* dat, int32 sampleFrames, int32 ch, int32 bitRate, bool sign, bool bigEndian)
{
	allSoundStop();
	sampleDatas[id].loadPCMData(dat, sampleFrames, ch, bitRate, sign, bigEndian);
}
void SynthesizerFM4::clearSampleData(int32 id)
{
	allSoundStop();
	sampleDatas[id].clear();
}
void SynthesizerFM4::factoryDataReset()
{
	if (_voiceUnitExt != nullptr) allSoundStop();
	// waveData
	const double _RES = double(-SynthData::LOG2VT_DYN_RES),
		_BOADER = 1.0 / (1 << SynthData::LOG2VT_DYN_BITS);
	const int _BOTTOM = SynthData::LOG2VT_SIZE;
	const double _PID2 = 3.141592653589793 * 2.0;
	const int _SIZE = WAVE_TABLE_SIZE, _HALF = _SIZE >> 1, _QTER = _SIZE >> 2;
	const double _SIZEd = double(_SIZE), _HALFd = double(_HALF), _QTERd = double(_QTER);
	double _sl[OVER_TONES]{0.0};
	double _sw[OVER_TONES]{0.0};
	
	if (waveTables == nullptr)waveTables = new int32[8 << WAVE_TABLE_SIZE_BITS];
	int32* _wave0 = waveTables;
	int32* _wave1 = &waveTables[1 << WAVE_TABLE_SIZE_BITS];
	int32* _wave2 = &waveTables[2 << WAVE_TABLE_SIZE_BITS];
	int32* _wave3 = &waveTables[3 << WAVE_TABLE_SIZE_BITS];
	int32* _wave4 = &waveTables[4 << WAVE_TABLE_SIZE_BITS];
	int32* _wave5 = &waveTables[5 << WAVE_TABLE_SIZE_BITS];
	int32* _wave6 = &waveTables[6 << WAVE_TABLE_SIZE_BITS];
	int32* _wave7 = &waveTables[7 << WAVE_TABLE_SIZE_BITS];
	for (int i = 0, _iv = 0; i < _HALF; ++i)
	{
	// sine [ nu ]
		double _nv = std::sin((i / _SIZEd) * _PID2);
		if      (_nv >  _BOADER) _iv =  int(std::log2( _nv) * _RES + 0.5) << 1;
		else if (_nv < -_BOADER) _iv = (int(std::log2(-_nv) * _RES + 0.5) << 1) + 1;
		else _iv = _BOTTOM;
		_wave0[i] = _iv;
		_wave0[_HALF + i] = _iv + 1;
	// sine [ n- ]
		_wave1[i] = _iv;
		_wave1[_HALF + i] = _BOTTOM;
	// sine [ nn ]
		_wave2[i] = _iv; 
		_wave2[_HALF + i] = _iv;
		if	((i & 1) == 0){ 
	// sine [nu-- ]
			_wave3[i >> 1] = _iv; 
			_wave3[_QTER + (i >> 1)] = _iv + 1;
	// sine [ nu-- ]
			_wave4[i >> 1] = _iv; 
			_wave4[_QTER + (i >> 1)] = _iv; }
		_wave3[_HALF + i] = _BOTTOM;
		_wave4[_HALF + i] = _BOTTOM;
	}
// triangle ----
	for (int i = 0, j = 1; i < OVER_TONES; ++i, j += 2)_sw[i] = _SIZEd / j;
	for (int i = 0, j = 1; i < OVER_TONES; ++i, j += 2)_sl[i] = (i & 1)? -1.0 / (j * j) : 1.0 / (j * j);
	double _den = 0.0;
	for (int j = 0; j < OVER_TONES; ++j)_den += std::sin(_QTERd / _sw[j] * _PID2) * _sl[j];
	_den = 1.0 / _den;
	for (int i = 0, _iv = 0; i < _SIZE; ++i)
	{
		double _nv = 0.0;
		for (int j = 0; j < OVER_TONES; ++j)_nv += std::sin((i / _sw[j]) * _PID2) * _sl[j];
		_nv *= _den;
		if      (_nv >  _BOADER) _iv =  int(std::log2( _nv) * _RES + 0.5) << 1;
		else if (_nv < -_BOADER) _iv = (int(std::log2(-_nv) * _RES + 0.5) << 1) + 1;
		else _iv = _BOTTOM;
		_wave5[i] = _iv;
	}
// square ----
	//for (int i = 0, j = 1; i < OVER_TONES; ++i, j += 2)_sw[i] = _SIZEd / j;
	for (int i = 0, j = 1; i < OVER_TONES; ++i, j += 2)_sl[i] = 1.0 / j;
	const double _cSQR = _QTERd / OVER_TONES;
	_den = 0.0;
	for (int j = 0; j < OVER_TONES; ++j)_den += std::sin(_cSQR / _sw[j] * _PID2) * _sl[j];
	_den = 1.0 / _den;
	for (int i = 0, _iv = 0; i < _SIZE; ++i)
	{
		double _nv = 0.0;
		for (int j = 0; j < OVER_TONES; ++j)_nv += std::sin((i / _sw[j]) * _PID2) * _sl[j];
		_nv *= _den;
		if      (_nv >  _BOADER) _iv =  int(std::log2( _nv) * _RES + 0.5) << 1;
		else if (_nv < -_BOADER) _iv = (int(std::log2(-_nv) * _RES + 0.5) << 1) + 1;
		else _iv = _BOTTOM;
		_wave6[i] = _iv;
	}
// saw ----
	for (int i = 0, j = 1; i < OVER_TONES; ++i, ++j)_sw[i] = _SIZE / j;
	for (int i = 0, j = 1; i < OVER_TONES; ++i, ++j)_sl[i] = 1.0 / j;
	double _c = _HALFd;
	for (int i = 0, j = 1; i < OVER_TONES; ++i, ++j) _c /= 1.0 + 1.0 / j;
	_c = std::round(_c + 0.5);
	_den = 0.0;
	for (int j = 0; j < OVER_TONES; ++j)_den += std::sin(_c / _sw[j] * _PID2) * _sl[j];
	_den = 1.0 / _den;
	for (int i = 0, _iv = 0; i < _SIZE; ++i)
	{
		double _nv = 0.0;
		for (int j = 0; j < OVER_TONES; ++j)_nv += std::sin((i / _sw[j]) * _PID2) * _sl[j];
		_nv *= _den;
		if      (_nv >  _BOADER) _iv =  int(std::log2( _nv) * _RES + 0.5) << 1;
		else if (_nv < -_BOADER) _iv = (int(std::log2(-_nv) * _RES + 0.5) << 1) + 1;
		else _iv = _BOTTOM;
		_wave7[i] = _iv;
	}
	// Preset
	if (tonePresets == nullptr)tonePresets = new TonePresetFM4[256];
	for (int i = 0; i < 256; ++i)tonePresets[i].reset();
// Sampler
	if (sampleDatas == nullptr)sampleDatas = new SampleData[128];
	if (samplerPresets == nullptr)samplerPresets = new SamplerPreset[128];
	for (int i = 0; i < 128; ++i)
	{
		sampleDatas[i].clear();
		samplerPresets[i].reset();
	}
}
// Constructor
SynthesizerFM4::SynthesizerFM4(int32 maxPoly, float refPitch, float masVol, bool privateTableData) : MultiTimberSynthBase::MultiTimberSynthBase(maxPoly, refPitch, masVol)
{
	_modelID = 0x22;
	_maxPMN = 3;
	// create
	_synthData = (privateTableData) ? _prvSynthData = new SynthData() : SynthData::getInstance();
	factoryDataReset();
	_recreateVU(_maxPolyphony);
	_resetKeyParams();
}
SynthesizerFM4::SynthesizerFM4(SynthesizerFM4 && x) noexcept : MultiTimberSynthBase::MultiTimberSynthBase(std::move(x))
{
	// Synthesizer - Data -----
	_prvSynthData = x._prvSynthData; x._prvSynthData = nullptr;
	_synthData = x._synthData; x._synthData = nullptr;
	waveTables = x.waveTables; x.waveTables = nullptr;
	sampleDatas = x.sampleDatas; x.sampleDatas = nullptr;
	tonePresets = x.tonePresets; x.tonePresets = nullptr;
	samplerPresets = x.samplerPresets; x.samplerPresets = nullptr;
	// Synthesizer - main -----
	_voiceUnitExt = x._voiceUnitExt; x._voiceUnitExt = nullptr;
}
SynthesizerFM4 & SynthesizerFM4::operator=(SynthesizerFM4 && x) noexcept
{
	if (this == &x)return *this;
	// MultiTimberSynthBase::operator=(std::move(x));
// MultiTimberSynthBase ===
	// Synth Status
	midiInPorts = x.midiInPorts;
	midiOutPorts = x.midiOutPorts;
	receiveSysEx = x.receiveSysEx;
	_manufacID = x._manufacID;
	_deiceID = x._deiceID;
	_modelID = x._modelID;
	// Channel Paramters
	if (voiceParts != nullptr)delete[] voiceParts;
	voiceParts = x.voiceParts; x.voiceParts = nullptr;
	_sampleRate = x._sampleRate;
	_referencePitch = x._referencePitch;
	_masterVolume = x._masterVolume;
	// VoiceUnits
	_maxPolyphony = x._maxPolyphony;
	if (_voiceUnits != nullptr)delete[] _voiceUnits;
	_voiceUnits = x._voiceUnits; x._voiceUnits = nullptr;
	_nton = x._nton;
	_ntonM = x._ntonM;
	// RPN,NRPN -----
	_rpnCtrlFlag = x._rpnCtrlFlag;
	_nrpnCtrlFlag = x._nrpnCtrlFlag;
// SynthesizerFM4 ===
	// Synthesizer - Data -----
	_synthData = x._synthData;
	if (_prvSynthData != nullptr)delete _prvSynthData;
	_prvSynthData = x._prvSynthData; x._prvSynthData = nullptr;
	if (waveTables != nullptr)delete[] waveTables;
	waveTables = x.waveTables; x.waveTables = nullptr;
	if (sampleDatas != nullptr)delete[] sampleDatas;
	sampleDatas = x.sampleDatas; x.sampleDatas = nullptr;
	if (tonePresets != nullptr)delete[] tonePresets;
	tonePresets = x.tonePresets; x.tonePresets = nullptr;
	if (samplerPresets != nullptr)delete[] samplerPresets;
	samplerPresets = x.samplerPresets; x.samplerPresets = nullptr;
	// Synthesizer - main -----
	if (_voiceUnitExt != nullptr)delete[] _voiceUnitExt;
	_voiceUnitExt = x._voiceUnitExt; x._voiceUnitExt = nullptr;

	return *this;
}
SynthesizerFM4::~SynthesizerFM4()
{
	if (voiceParts != nullptr)delete[] voiceParts;
	voiceParts = nullptr;
	if (_voiceUnits != nullptr)delete[] _voiceUnits;
	_voiceUnits = nullptr;
	if (waveTables != nullptr)delete[] waveTables;
	waveTables = nullptr;
	if (sampleDatas != nullptr)delete[] sampleDatas;
	sampleDatas = nullptr;
	if (tonePresets != nullptr)delete[] tonePresets;
	tonePresets = nullptr;
	if (samplerPresets != nullptr)delete[] samplerPresets;
	samplerPresets = nullptr;
	if (_voiceUnitExt != nullptr)delete[] _voiceUnitExt;
	_voiceUnitExt = nullptr;
	if (_prvSynthData != nullptr)delete _prvSynthData;
	_prvSynthData = nullptr;
}
// MidiEvent ---
void SynthesizerFM4::_noteOn(const MidiEvent& ev)
{
	VoicePart& _pt = voiceParts[ev.channel];
	for (int i = 0; i < _maxPolyphony; ++i)
	{
		VoiceUnitFM4& _vu = _voiceUnitExt[i];
		if (_vu.state == 0)
		{
			_vu.startFrame = ev.offsetFrames;
			if (_pt.mode == 3)_vu.keyOnSampler(ev.channel, ev.value0, ev.value1);
			else _vu.keyOn(ev.channel, ev.value0, ev.value1);
			return;
		}
	}
	const int _len = _nton + _maxPolyphony;
	for (int i = _nton; i < _len; ++i)
	{
		VoiceUnitFM4& _vu = _voiceUnitExt[i & _ntonM];
		// (KeyOff || SoundOff) -->
		if (_vu.state & 0x60)
		{
			_nton = _nton + 1 & _ntonM;
			_vu.startFrame = ev.offsetFrames;
			if (_pt.mode == 3)_vu.keyOnSampler(ev.channel, ev.value0, ev.value1);
			else _vu.keyOn(ev.channel, ev.value0, ev.value1);
			return;
		}
	}
	for (int i = _nton; i < _len; ++i)
	{
		VoiceUnitFM4& _vu = _voiceUnitExt[i & _ntonM];
		// KeyOn -->
		if (_vu.state & 0x10)
		{
			_nton = _nton + 1 & _ntonM;
			_vu.startFrame = ev.offsetFrames;
			if (_pt.mode == 3)_vu.keyOnSampler(ev.channel, ev.value0, ev.value1);
			else _vu.keyOn(ev.channel, ev.value0, ev.value1);
			return;
		}
	}
}
void SynthesizerFM4::_readSysExData(const MidiEvent& ev)
{
	if (ev.dataSize < 8)return;
	const uint8* dat = ev.data;
	const uint8* dataEnd = dat + ev.dataSize - 1;  // minus F7
	if (*(dat++) != _manufacID)return;
	if (*(dat++) != _deiceID)return;
	if (*(dat++) != _modelID)return;

	uint8 _padr = 0u;
	switch (*(dat++))	// Address 1
	{
	// Common
		case 0x00u:
			if (*(dat++) != 0x00u)return; // Address 2
			switch (*(dat++))	// Address 3
			{
				case 0x00u : setMasterVolume(_masterVolumeTable[*(dat++)]); if (dat >= dataEnd) break;
				case 0x01u : setReferencePitch(440.0f + (*dat - 64.0f) * 0.5f); break;
			}
			break;
	// Part
		case 0x02u: {
			int32 v = *(dat++) & 15;	// Address 2 [ CH ]
			switch (*(dat++))	// Address 3
			{
				case 0x00u : setDefaultPartMode(v, *(dat++)); if (dat >= dataEnd) break;
				case 0x01u : setPartMode(v, *(dat++)); break;
			}
			} break;
	// Inst (Drum map)
		case 0x05u: _padr = 128u;
	// Inst (Program map)
		case 0x04u:{
			TonePresetFM4& _preset = tonePresets[_padr | *(dat++)];	// Address 2 [ ID ]
			switch (*(dat++))	// Address 3
			{
				case 0x00u : _preset.toneName[ 0] = *(dat++);  if (dat >= dataEnd) break;
				case 0x01u : _preset.toneName[ 1] = *(dat++);  if (dat >= dataEnd) break;
				case 0x02u : _preset.toneName[ 2] = *(dat++);  if (dat >= dataEnd) break;
				case 0x03u : _preset.toneName[ 3] = *(dat++);  if (dat >= dataEnd) break;
				case 0x04u : _preset.toneName[ 4] = *(dat++);  if (dat >= dataEnd) break;
				case 0x05u : _preset.toneName[ 5] = *(dat++);  if (dat >= dataEnd) break;
				case 0x06u : _preset.toneName[ 6] = *(dat++);  if (dat >= dataEnd) break;
				case 0x07u : _preset.toneName[ 7] = *(dat++);  if (dat >= dataEnd) break;
				case 0x08u : _preset.toneName[ 8] = *(dat++);  if (dat >= dataEnd) break;
				case 0x09u : _preset.toneName[ 9] = *(dat++);  if (dat >= dataEnd) break;
				case 0x0Au : _preset.toneName[10] = *(dat++);  if (dat >= dataEnd) break;
				case 0x0Bu : _preset.toneName[11] = *(dat++);  if (dat >= dataEnd) break;
				case 0x0Cu : _preset.toneName[12] = *(dat++);  if (dat >= dataEnd) break;
				case 0x0Du : _preset.toneName[13] = *(dat++);  if (dat >= dataEnd) break;
				case 0x0Eu : _preset.toneName[14] = *(dat++);  if (dat >= dataEnd) break;
				case 0x0Fu : _preset.toneName[15] = *(dat++);  if (dat >= dataEnd) break;
				
				case 0x10u : ++dat; if (dat >= dataEnd) break;
				case 0x11u : ++dat; if (dat >= dataEnd) break;
				case 0x12u : ++dat; if (dat >= dataEnd) break;
				case 0x13u : ++dat; if (dat >= dataEnd) break;
				case 0x14u : ++dat; if (dat >= dataEnd) break;
				case 0x15u : ++dat; if (dat >= dataEnd) break;
				case 0x16u : ++dat; if (dat >= dataEnd) break;
				case 0x17u : ++dat; if (dat >= dataEnd) break;
				case 0x18u : ++dat; if (dat >= dataEnd) break;
				case 0x19u : _preset.key     = *(dat++);     if (dat >= dataEnd) break;
				case 0x1Au : _preset.pan     = *(dat++);     if (dat >= dataEnd) break;
				case 0x1Bu : _preset.lfoWave = *(dat++) & 3; if (dat >= dataEnd) break;
				case 0x1Cu : _preset.lfoFreq = _preset.lfoFreq & 0x3f80 | *(dat++);      if (dat >= dataEnd) break;
				case 0x1Du : _preset.lfoFreq = _preset.lfoFreq & 0x007f | *(dat++) << 7; if (dat >= dataEnd) break;
				case 0x1Eu : _preset.fl      = *(dat++);     if (dat >= dataEnd) break;
				case 0x1Fu : _preset.strc    = *(dat++) & 7; if (dat >= dataEnd) break;
				
				case 0x20u : ++dat; if (dat >= dataEnd) break;
				case 0x21u : ++dat; if (dat >= dataEnd) break;
				case 0x22u : ++dat; if (dat >= dataEnd) break;
				case 0x23u : _preset.tlVelCtrl_0 = *(dat++) & 7;  if (dat >= dataEnd) break;
				case 0x24u : _preset.mul_0       = *(dat++) & 15; if (dat >= dataEnd) break;
				case 0x25u : _preset.tune_0      = _preset.tune_0 & 0x3f80 | *(dat++);      if (dat >= dataEnd) break;
				case 0x26u : _preset.tune_0      = _preset.tune_0 & 0x007f | *(dat++) << 7; if (dat >= dataEnd) break;
				case 0x27u : _preset.lms_0       = *(dat++) & 15; if (dat >= dataEnd) break;
				case 0x28u : _preset.wave_0      = *(dat++) & 7;  if (dat >= dataEnd) break;
				case 0x29u : _preset.tl_0        = *(dat++);      if (dat >= dataEnd) break;
				case 0x2Au : _preset.ar_0        = *(dat++);      if (dat >= dataEnd) break;
				case 0x2Bu : _preset.dr_0        = *(dat++);      if (dat >= dataEnd) break;
				case 0x2Cu : _preset.sl_0        = *(dat++);      if (dat >= dataEnd) break;
				case 0x2Du : _preset.rr_0        = *(dat++);      if (dat >= dataEnd) break;
				case 0x2Eu : _preset.kf_0        = *(dat++) & 3;  if (dat >= dataEnd) break;
				case 0x2Fu : _preset.kfStart_0   = *(dat++);      if (dat >= dataEnd) break;
				
				case 0x30u : ++dat; if (dat >= dataEnd) break;
				case 0x31u : ++dat; if (dat >= dataEnd) break;
				case 0x32u : ++dat; if (dat >= dataEnd) break;
				case 0x33u : _preset.tlVelCtrl_1 = *(dat++) & 7;  if (dat >= dataEnd) break;
				case 0x34u : _preset.mul_1       = *(dat++) & 15; if (dat >= dataEnd) break;
				case 0x35u : _preset.tune_1      = _preset.tune_1 & 0x3f80 | *(dat++);      if (dat >= dataEnd) break;
				case 0x36u : _preset.tune_1      = _preset.tune_1 & 0x007f | *(dat++) << 7; if (dat >= dataEnd) break;
				case 0x37u : _preset.lms_1       = *(dat++) & 15; if (dat >= dataEnd) break;
				case 0x38u : _preset.wave_1      = *(dat++) & 7;  if (dat >= dataEnd) break;
				case 0x39u : _preset.tl_1        = *(dat++);      if (dat >= dataEnd) break;
				case 0x3Au : _preset.ar_1        = *(dat++);      if (dat >= dataEnd) break;
				case 0x3Bu : _preset.dr_1        = *(dat++);      if (dat >= dataEnd) break;
				case 0x3Cu : _preset.sl_1        = *(dat++);      if (dat >= dataEnd) break;
				case 0x3Du : _preset.rr_1        = *(dat++);      if (dat >= dataEnd) break;
				case 0x3Eu : _preset.kf_1        = *(dat++) & 3;  if (dat >= dataEnd) break;
				case 0x3Fu : _preset.kfStart_1   = *(dat++);      if (dat >= dataEnd) break;

				case 0x40u : ++dat; if (dat >= dataEnd) break;
				case 0x41u : ++dat; if (dat >= dataEnd) break;
				case 0x42u : ++dat; if (dat >= dataEnd) break;
				case 0x43u : _preset.tlVelCtrl_2 = *(dat++) & 7;  if (dat >= dataEnd) break;
				case 0x44u : _preset.mul_2       = *(dat++) & 15; if (dat >= dataEnd) break;
				case 0x45u : _preset.tune_2      = _preset.tune_2 & 0x3f80 | *(dat++);      if (dat >= dataEnd) break;
				case 0x46u : _preset.tune_2      = _preset.tune_2 & 0x007f | *(dat++) << 7; if (dat >= dataEnd) break;
				case 0x47u : _preset.lms_2       = *(dat++) & 15; if (dat >= dataEnd) break;
				case 0x48u : _preset.wave_2      = *(dat++) & 7;  if (dat >= dataEnd) break;
				case 0x49u : _preset.tl_2        = *(dat++);      if (dat >= dataEnd) break;
				case 0x4Au : _preset.ar_2        = *(dat++);      if (dat >= dataEnd) break;
				case 0x4Bu : _preset.dr_2        = *(dat++);      if (dat >= dataEnd) break;
				case 0x4Cu : _preset.sl_2        = *(dat++);      if (dat >= dataEnd) break;
				case 0x4Du : _preset.rr_2        = *(dat++);      if (dat >= dataEnd) break;
				case 0x4Eu : _preset.kf_2        = *(dat++) & 3;  if (dat >= dataEnd) break;
				case 0x4Fu : _preset.kfStart_2   = *(dat++);      if (dat >= dataEnd) break;

				case 0x50u : ++dat; if (dat >= dataEnd) break;
				case 0x51u : ++dat; if (dat >= dataEnd) break;
				case 0x52u : ++dat; if (dat >= dataEnd) break;
				case 0x53u : _preset.tlVelCtrl_3 = *(dat++) & 7;  if (dat >= dataEnd) break;
				case 0x54u : _preset.mul_3       = *(dat++) & 15; if (dat >= dataEnd) break;
				case 0x55u : _preset.tune_3      = _preset.tune_3 & 0x3f80 | *(dat++);      if (dat >= dataEnd) break;
				case 0x56u : _preset.tune_3      = _preset.tune_3 & 0x007f | *(dat++) << 7; if (dat >= dataEnd) break;
				case 0x57u : _preset.lms_3       = *(dat++) & 15; if (dat >= dataEnd) break;
				case 0x58u : _preset.wave_3      = *(dat++) & 7;  if (dat >= dataEnd) break;
				case 0x59u : _preset.tl_3        = *(dat++);      if (dat >= dataEnd) break;
				case 0x5Au : _preset.ar_3        = *(dat++);      if (dat >= dataEnd) break;
				case 0x5Bu : _preset.dr_3        = *(dat++);      if (dat >= dataEnd) break;
				case 0x5Cu : _preset.sl_3        = *(dat++);      if (dat >= dataEnd) break;
				case 0x5Du : _preset.rr_3        = *(dat++);      if (dat >= dataEnd) break;
				case 0x5Eu : _preset.kf_3        = *(dat++) & 3;  if (dat >= dataEnd) break;
				case 0x5Fu : _preset.kfStart_3   = *(dat++);      break;
			}
			} break;
	// Percussion
		case 0x06u: {
			SamplerPreset& _smplPrst = samplerPresets[*(dat++)];	// Address 2 [ ID ]
			switch (*(dat++))	// Address 3
			{
				case 0x00u : _smplPrst.toneName[ 0] = *(dat++);  if (dat >= dataEnd) break;
				case 0x01u : _smplPrst.toneName[ 1] = *(dat++);  if (dat >= dataEnd) break;
				case 0x02u : _smplPrst.toneName[ 2] = *(dat++);  if (dat >= dataEnd) break;
				case 0x03u : _smplPrst.toneName[ 3] = *(dat++);  if (dat >= dataEnd) break;
				case 0x04u : _smplPrst.toneName[ 4] = *(dat++);  if (dat >= dataEnd) break;
				case 0x05u : _smplPrst.toneName[ 5] = *(dat++);  if (dat >= dataEnd) break;
				case 0x06u : _smplPrst.toneName[ 6] = *(dat++);  if (dat >= dataEnd) break;
				case 0x07u : _smplPrst.toneName[ 7] = *(dat++);  if (dat >= dataEnd) break;
				case 0x08u : _smplPrst.toneName[ 8] = *(dat++);  if (dat >= dataEnd) break;
				case 0x09u : _smplPrst.toneName[ 9] = *(dat++);  if (dat >= dataEnd) break;
				case 0x0Au : _smplPrst.toneName[10] = *(dat++);  if (dat >= dataEnd) break;
				case 0x0Bu : _smplPrst.toneName[11] = *(dat++);  if (dat >= dataEnd) break;
				case 0x0Cu : _smplPrst.toneName[12] = *(dat++);  if (dat >= dataEnd) break;
				case 0x0Du : _smplPrst.toneName[13] = *(dat++);  if (dat >= dataEnd) break;
				case 0x0Eu : _smplPrst.toneName[14] = *(dat++);  if (dat >= dataEnd) break;
				case 0x0Fu : _smplPrst.toneName[15] = *(dat++);  if (dat >= dataEnd) break;

				case 0x10u : _smplPrst.smplNum = *(dat++);       if (dat >= dataEnd) break;
				case 0x11u : _smplPrst.pan     = *(dat++);       if (dat >= dataEnd) break;
				case 0x12u : _smplPrst.octv    = *(dat++) & 15;  if (dat >= dataEnd) break;
				case 0x13u : _smplPrst.tune    = _smplPrst.tune & 0x3f80 | *(dat++);      if (dat >= dataEnd) break;
				case 0x14u : _smplPrst.tune    = _smplPrst.tune & 0x007f | *(dat++) << 7; if (dat >= dataEnd) break;
				case 0x15u : _smplPrst.tl      = *(dat++);       if (dat >= dataEnd) break;
				case 0x16u : _smplPrst.ar      = *(dat++);       if (dat >= dataEnd) break;
				case 0x17u : _smplPrst.dr      = *(dat++);       if (dat >= dataEnd) break;
				case 0x18u : _smplPrst.sl      = *(dat++);       if (dat >= dataEnd) break;
				case 0x19u : _smplPrst.rr      = *dat;           break;
			}
			} break;
	}
}
/***********************************************
	Class	SynthesizerFM4 :: VoiceUnitFM4
************************************************/
void SynthesizerFM4::VoiceUnitFM4::keyOff()
{
	// KeyOn -->
	if (state & 0x0010)
	{
	// OP0 is Active -->
		if (state & 0b0001)
		{
			if (_egState_0 == 1)
			{
				_egLevelTable_0 = _EG_LEVEL_TABLE_DW;
				if (_egCount_0 >= _EG_MAX_CNT)_egCount_0 = 0;
				else _egCount_0 = _EG_LEVEL_TABLE_UP[_egCount_0 >> _EG_CNT_FP] << (_EG_CNT_FP - 1);  // -1 = 8192 to 4096
			}
			_egGate_0 = _egCntEnvBtm_0;
			_egCountStep_0 = _egCountStepRR_0;
			_egState_0 = 8;
			if (_egCount_0 > _egCntEnvBtm_0)
			{ 
				if (state & 0b1110){ state &= 0b11111110; _egState_0 = 0; }
				else { killProcess(); return; }
			}
			// sampler ---
			_loopE = _plbkE;
			_loopF = _loopFk;
		}
		// OP1 is Active -->
		if (state & 0b0010)
		{
			if (_egState_1 == 1)
			{
				_egLevelTable_1 = _EG_LEVEL_TABLE_DW;
				if (_egCount_1 >= _EG_MAX_CNT)_egCount_1 = 0;
				else _egCount_1 = _EG_LEVEL_TABLE_UP[_egCount_1 >> _EG_CNT_FP] << (_EG_CNT_FP - 1);  // -1 = 8192 to 4096
			}
			_egGate_1 = _egCntEnvBtm_1;
			_egCountStep_1 = _egCountStepRR_1;
			_egState_1 = 8;
			if (_egCount_1 > _egCntEnvBtm_1)
			{
				if (state & 0b1101) { state &= 0b11111101; _egState_1 = 0; }
				else { killProcess(); return; }
			}
		}
		// OP2 is Active -->
		if (state & 0b0100)
		{
			if (_egState_2 == 1)
			{
				_egLevelTable_2 = _EG_LEVEL_TABLE_DW;
				if (_egCount_2 >= _EG_MAX_CNT)_egCount_2 = 0;
				else _egCount_2 = _EG_LEVEL_TABLE_UP[_egCount_2 >> _EG_CNT_FP] << (_EG_CNT_FP - 1);  // -1 = 8192 to 4096
			}
			_egGate_2 = _egCntEnvBtm_2;
			_egCountStep_2 = _egCountStepRR_2;
			_egState_2 = 8;
			if (_egCount_2 > _egCntEnvBtm_2)
			{
				if (state & 0b1011) { state &= 0b11111011; _egState_2 = 0; }
				else { killProcess(); return; }
			}
		}
		// OP3 is Active -->
		if (state & 0b1000)
		{
			if (_egState_3 == 1)
			{
				_egLevelTable_3 = _EG_LEVEL_TABLE_DW;
				if (_egCount_3 >= _EG_MAX_CNT)_egCount_3 = 0;
				else _egCount_3 = _EG_LEVEL_TABLE_UP[_egCount_3 >> _EG_CNT_FP] << (_EG_CNT_FP - 1);  // -1 = 8192 to 4096
			}
			_egGate_3 = _egCntEnvBtm_3;
			_egCountStep_3 = _egCountStepRR_3;
			_egState_3 = 8;
			if (_egCount_3 > _egCntEnvBtm_3)
			{
				if (state & 0b0111) { state &= 0b11110111; _egState_3 = 0; }
				else { killProcess(); return; }
			}
		}
		state = 0x0020 | state & 0b1111;
	}
}
void SynthesizerFM4::VoiceUnitFM4::soundOff()
{
	// KeyOn || KeyOff --->
	if (state & 0x0030)
	{
	// OP0 Active --->
		if (state & 0b0001)
		{
			if (_egState_0 == 1)
			{
				_egLevelTable_0 = _EG_LEVEL_TABLE_DW;
				if (_egCount_0 >= _EG_MAX_CNT)_egCount_0 = 0;
				else _egCount_0 = _EG_LEVEL_TABLE_UP[_egCount_0 >> _EG_CNT_FP] << (_EG_CNT_FP - 1);  // -1 = 8192 to 4096
			}
			if (_egCount_0 > _egCntEnvBtm_0)
			{ 
				if (state & 0b1110){ state &= 0b11111110; }
				else { killProcess(); return; }
			}
			_egGate_0 = _egCntEnvBtm_0;
			_egCountStep_0 = egPhaseStepSndOff;
			_egState_0 = 0;
			// sampler ---
			_loopE = _plbkE;
			_loopF = _loopFk;
		}
	// OP1 Active --->
		if (state & 0b0010)
		{
			if (_egState_1 == 1)
			{
				_egLevelTable_1 = _EG_LEVEL_TABLE_DW;
				if (_egCount_1 >= _EG_MAX_CNT)_egCount_1 = 0;
				else _egCount_1 = _EG_LEVEL_TABLE_UP[_egCount_1 >> _EG_CNT_FP] << (_EG_CNT_FP - 1);  // -1 = 8192 to 4096
			}
			if (_egCount_1 > _egCntEnvBtm_1)
			{ 
				if (state & 0b1101){ state &= 0b11111101; }
				else { killProcess(); return; }
			}
			_egGate_1 = _egCntEnvBtm_1;
			_egCountStep_1 = egPhaseStepSndOff;
			_egState_1 = 0;
		}
	// OP2 Active --->
		if (state & 0b0100)
		{
			if (_egState_2 == 1)
			{
				_egLevelTable_2 = _EG_LEVEL_TABLE_DW;
				if (_egCount_2 >= _EG_MAX_CNT)_egCount_2 = 0;
				else _egCount_2 = _EG_LEVEL_TABLE_UP[_egCount_2 >> _EG_CNT_FP] << (_EG_CNT_FP - 1);  // -1 = 8192 to 4096
			}
			if (_egCount_2 > _egCntEnvBtm_2)
			{ 
				if (state & 0b1011){ state &= 0b11111011; }
				else { killProcess(); return; }
			}
			_egGate_2 = _egCntEnvBtm_2;
			_egCountStep_2 = egPhaseStepSndOff;
			_egState_2 = 0;
			// sampler ---
			_loopE = _plbkE;
			_loopF = _loopFk;
		}
	// OP3 Active --->
		if (state & 0b1000)
		{
			if (_egState_3 == 1)
			{
				_egLevelTable_3 = _EG_LEVEL_TABLE_DW;
				if (_egCount_3 >= _EG_MAX_CNT)_egCount_3 = 0;
				else _egCount_3 = _EG_LEVEL_TABLE_UP[_egCount_3 >> _EG_CNT_FP] << (_EG_CNT_FP - 1);  // -1 = 8192 to 4096
			}
			if (_egCount_3 > _egCntEnvBtm_3)
			{
				if (state & 0b0111) { state &= 0b11110111; }
				else { killProcess(); return; }
			}
			_egGate_3 = _egCntEnvBtm_3;
			_egCountStep_3 = egPhaseStepSndOff;
			_egState_3 = 0;
			// sampler ---
			_loopE = _plbkE;
			_loopF = _loopFk;
		}
		state = 0x0040 | state & 0b1111;
		partNumber = -1; // SoundOff
	}
}

int32 SynthesizerFM4::VoiceUnitFM4::_shiftEGState(int32 op)
{
	switch (_egState[op])
	{
		// ATTACK ->
		case 1:
			_egLevelTable[op] = _EG_LEVEL_TABLE_DW;
			_egCount[op] = 0;
			// -> DECAY
			if (_egSL[op] > 0)
			{
				_egGate[op] = _egCntEnvSus[op];
				_egCountStep[op] = _egCountStepDR[op];
				return _egState[op] = 2;
			}
			// -> SUSTAIN
			_egGate[op] = 0x7fffffff;
			_egCountStep[op] = 0;
			return _egState[op] = 4;
		// DECAY ->
		case 2:
			// -> SUSTAIN
			if ((_egTL[op] + _egSL[op]) < _EG_SUSTAIN_BTM)
			{
				_egLevelTable[op] = _EG_LEVEL_TABLE_DW;
				_egGate[op] = 0x7fffffff;
				_egCountStep[op] = 0;
				_egCount[op] = _egCntEnvSus[op];
				return _egState[op] = 4;
			}
		// DECAY,RELEASE,SoundOFF -> EG OFF
		default: return _egState[op] = 0;
	}
}

void SynthesizerFM4::VoiceUnitFM4::keyOn(const int32& ch, const int32& note, const int32& velo)
{
// KeyOn Status ---
	state = 0b00011111;
	partNumber = ch;
	noteNumber = note;
// tone ---
	VoicePart& _part = VOICE_PARTS[partNumber];
	TonePresetFM4* _preset = nullptr;
	int _keyFlw[4] = { egPow2tIdx, egPow2tIdx, egPow2tIdx, egPow2tIdx };
	if (_part.mode == 2)
	{
		_preset = &TONE_PRESETS[noteNumber | 128];
		for (int i = 0; i < 4; ++i)_pgKey[i] = _POW2T_KEY_TABLE[_preset->key] + _MUL_TABLE[_preset->mul[i]] + (_preset->tune[i] - 8192) + pgKeyPow2tIdx;
		//_keyFlw = 0;
		_prPan = _preset->pan;
	}
	else
	{
		_preset = &TONE_PRESETS[_part.programChange];
		for (int i = 0; i < 4; ++i)
		{
			_pgKey[i] = _POW2T_KEY_TABLE[noteNumber] + _MUL_TABLE[_preset->mul[i]] + (_preset->tune[i] - 8192) + pgKeyPow2tIdx;
			_keyFlw[i] += _EG_KEY_TABLE[127 + noteNumber - _preset->kfStart[i]] << _preset->kf[i];
		}
		_prPan = 64;
	}
// wave ---
	_fdbk = _EG_TOTAL_LEVEL_TABLE[_preset->fl];
// LFO ---
	_lfoWaveTable = &_LFO_WAVE_TABLES[_preset->lfoWave << _LFO_WAVE_TABLE_SIZE_BITS];
	_lfoTimerStep = _preset->lfoFreq * lfoTimerCoef;
	_lfoTimer = _LFO_TIMER_INIT_VALUE;
	_lfoPhase = _preset->strc >> 3;  // ~8 AM 
// volume ---
	_vPtVol = _mainVolTable[_part.ccValues[7]];
	_vPtExp = _exprVolTable[_part.ccValues[11]];
	_vVel = _veloVolTable[velo];
	_vVol = _vVel * _masterVolume;
	_vOut = _vPtVol * _vPtExp * _vVol;
	_vPanL = _panTable[      _prPan + _part.ccValues[10]];
	_vPanR = _panTable[256 - _prPan - _part.ccValues[10]];
	_vOutL = _vOut * _vPanL;
	_vOutR = _vOut * _vPanR;
// pitch ---
	_ptBendSens = _POW2T_KEY_TABLE[_part.bendSens];
	_ptBend = _part.pitchBend - 8192;
	_pgBend = _ptBend * _ptBendSens >> 13;
// process ---
	process = _PROC_PTRS[_preset->strc];
// Operator --->
	for (int i = 0; i < 4; ++i)
	{
		// Operator 0 --->
		_waveTable[i] = &WAVE_TABLES[_preset->wave[i] << _WAVE_TABLE_SIZE_BITS];
		// LFO ---
		_lfoModSens[i] = _preset->lms[i];
		_lfoMod[i] = _part.ccValues[1] * _lfoModSens[i];
		_lfoOutPM[i] = 0;
		_lfoOutAM[i] = 0;
		// PG ---
		int v = _pgKey[i] + _pgBend;
		_pgPhaseStep[i] = _POW2_TABLE[v & _POW2T_OCTV_MASK] >> _POW2T_RS_TABLE[v >> _POW2T_OCTV_BITS];
		_pgPhase[i] = 0;
		// EG ---
		_egTL[i] = _preset->tl[i] + ((127 - velo) >> (7 - _preset->tlVelCtrl[i]));
		if (_egTL[i] > 127)_egTL[i] = 0;
		_egTLv[i] = _EG_TOTAL_LEVEL_TABLE[_egTL[i]];
		_egCntEnvBtm[i] = _EG_CNT_GATE_TABLE[128 - _egTL[i]];
		_egSL[i] = _preset->sl[i];
		_egCntEnvSus[i] = _EG_CNT_GATE_TABLE[_egSL[i]];
		if (_egCntEnvSus[i] > _egCntEnvBtm[i])_egCntEnvSus[i] = _egCntEnvBtm[i];
		v = (_preset->dr[i] << _EG_PRAM_OCTV_LS) + _keyFlw[i];
		_egCountStepDR[i] = _POW2_TABLE[v & _POW2T_OCTV_MASK] >> _EG_POW2T_RS_TABLE[v >> _POW2T_OCTV_BITS];
		v = (_preset->rr[i] << _EG_PRAM_OCTV_LS) + _keyFlw[i];
		_egCountStepRR[i] = _POW2_TABLE[v & _POW2T_OCTV_MASK] >> _EG_POW2T_RS_TABLE[v >> _POW2T_OCTV_BITS];
		// EG State ---
			// -> ATTACK
		if (_preset->ar[i] < 127) {
			_egCount[i] = _EG_CNT_ENV_ATK;
			_egLevelTable[i] = _EG_LEVEL_TABLE_UP;
			_egGate[i] = _EG_MAX_CNT;
			v = (_preset->ar[i] << _EG_PRAM_OCTV_LS) + _keyFlw[i];
			_egCountStep[i] = _POW2_TABLE[v & _POW2T_OCTV_MASK] >> _EG_POW2T_RS_TABLE[v >> _POW2T_OCTV_BITS];
			_egState[i] = 1;
		}
		else {
			_egLevelTable[i] = _EG_LEVEL_TABLE_DW;
			_egCount[i] = 0;
			// -> DECAY
			if (_egSL[i] > 0) {
				_egGate[i] = _egCntEnvSus[i];
				_egCountStep[i] = _egCountStepDR[i];
				_egState[i] = 2;
				// -> SUSTAIN
			}
			else {
				_egGate[i] = 0x7fffffff;
				_egCountStep[i] = 0;
				_egState[i] = 4;
			}
		}
	}
}
void SynthesizerFM4::VoiceUnitFM4::keyOnSampler(const int32& ch, const int32& note, const int32& velo)
{
// KeyOn Status ---
	state = 0b00010001;
	partNumber = ch;
	noteNumber = note;
// tone ---
	VoicePart& _part = VOICE_PARTS[partNumber];
	SamplerPreset& _preset = SAMPLER_PRESETS[noteNumber];
	SampleData& _sample = SAMPLE_DATAS[_preset.smplNum];
	if (_sample.sampleData == nullptr){ killProcess(); return; }
	_waveTable_0 = _sample.sampleData;
	_plbkE = _sample.plbkE;
	_loopE = _sample.loopE;
	_loopF = _sample.loopF;
	_loopFk = _sample.loopFk;
// LFO ---
	_lfoOutPM_0 = 0;
	_lfoOutAM_0 = 0;
// volume ---
	_vPtVol = _mainVolTable[_part.ccValues[7]];
	_vPtExp = _exprVolTable[_part.ccValues[11]];
	_vVel = _veloVolTable[velo];
	_vVol = _vVel * _masterVolume;
	_vOut = _vPtVol * _vPtExp * _vVol;
	_prPan = _preset.pan;
	_vPanL = _panTable[      _prPan + _part.ccValues[10]];
	_vPanR = _panTable[256 - _prPan - _part.ccValues[10]];
	_vOutL = _vOut * _vPanL;
	_vOutR = _vOut * _vPanR;
// PG ---
	_ptBendSens = _POW2T_KEY_TABLE[_part.bendSens];
	_ptBend = _part.pitchBend - 8192;
	_pgBend = _ptBend * _ptBendSens >> 13;
	_pgKey_0 = ((_preset.octv - 8) << 13) + (_preset.tune - 8192) + _sample.pgSample + pgSamplePow2tIdx;
	int v = _pgKey_0 + _pgBend;
	_pgPhaseStep_0 = _POW2_TABLE[v & _POW2T_OCTV_MASK] >> _POW2T_RS_TABLE[v >> _POW2T_OCTV_BITS];
	_pgPhase_0 = _sample.plbkS;
	_pgPhaseAFP = 0;
// EG ---
	_egTL_0 = _preset.tl;
	_egTLv_0 = _EG_TOTAL_LEVEL_TABLE[_egTL_0];
	_egCntEnvBtm_0 = _EG_CNT_GATE_TABLE[128 - _egTL_0];
	_egSL_0 = _preset.sl;
	_egCntEnvSus_0 = _EG_CNT_GATE_TABLE[_egSL_0];
	if (_egCntEnvSus_0 > _egCntEnvBtm_0)_egCntEnvSus_0 = _egCntEnvBtm_0;
	v = (_preset.dr << _EG_PRAM_OCTV_LS) + egPow2tIdx;
	_egCountStepDR_0 = _POW2_TABLE[v & _POW2T_OCTV_MASK] >> _EG_POW2T_RS_TABLE[v >> _POW2T_OCTV_BITS];
	v = (_preset.rr << _EG_PRAM_OCTV_LS) + egPow2tIdx;
	_egCountStepRR_0 = _POW2_TABLE[v & _POW2T_OCTV_MASK] >> _EG_POW2T_RS_TABLE[v >> _POW2T_OCTV_BITS];
// EG State ---
	// -> ATTACK
	if (_preset.ar < 127) {
		_egCount_0 = _EG_CNT_ENV_ATK;
		_egLevelTable_0 = _EG_LEVEL_TABLE_UP;
		_egGate_0 = _EG_MAX_CNT;
		v = (_preset.ar << _EG_PRAM_OCTV_LS) + egPow2tIdx;
		_egCountStep_0 = _POW2_TABLE[v & _POW2T_OCTV_MASK] >> _EG_POW2T_RS_TABLE[v >> _POW2T_OCTV_BITS];
		_egState_0 = 1;
	} else {
		_egLevelTable_0 = _EG_LEVEL_TABLE_DW;
		_egCount_0 = 0;
	// -> DECAY
		if (_egSL_0 > 0) {
			_egGate_0 = _egCntEnvSus_0;
			_egCountStep_0 = _egCountStepDR_0;
			_egState_0 = 2;
	// -> SUSTAIN
		} else {
			_egGate_0 = 0x7fffffff;
			_egCountStep_0 = 0;
			_egState_0 = 4; }
	}
// process ---
	process = _PROC_PTRS_SMPL[(_sample.channelCount-1) & 1];
}
void SynthesizerFM4::VoiceUnitFM4::updPartBendSens(const int32& val)
{
	_ptBendSens = _POW2T_KEY_TABLE[val];
	_pgBend = _ptBend * _ptBendSens >> 13;
	int v;
	// Inst op0 --->
	if (state & 0b0001)
	{
		v = _pgKey_0 + _pgBend + _lfoOutPM_0;
		_pgPhaseStep_0 = _POW2_TABLE[v & _POW2T_OCTV_MASK] >> _POW2T_RS_TABLE[v >> _POW2T_OCTV_BITS];
	}
	// Inst op1 --->
	if (state & 0b0010)
	{
		v = _pgKey_1 + _pgBend + _lfoOutPM_1;
		_pgPhaseStep_1 = _POW2_TABLE[v & _POW2T_OCTV_MASK] >> _POW2T_RS_TABLE[v >> _POW2T_OCTV_BITS];
	}
	// Inst op2 --->
	if (state & 0b0100)
	{
		v = _pgKey_2 + _pgBend + _lfoOutPM_2;
		_pgPhaseStep_2 = _POW2_TABLE[v & _POW2T_OCTV_MASK] >> _POW2T_RS_TABLE[v >> _POW2T_OCTV_BITS];
	}
	// Inst op3 --->
	if (state & 0b1000)
	{
		v = _pgKey_3 + _pgBend + _lfoOutPM_3;
		_pgPhaseStep_3 = _POW2_TABLE[v & _POW2T_OCTV_MASK] >> _POW2T_RS_TABLE[v >> _POW2T_OCTV_BITS];
	}
}
void SynthesizerFM4::VoiceUnitFM4::updPartPitchBend(const int32& val)
{
	_ptBend = val - 8192;
	_pgBend = _ptBend * _ptBendSens >> 13;
	int v;
	// Inst op0 --->
	if (state & 0b0001)
	{
		v = _pgKey_0 + _pgBend + _lfoOutPM_0;
		_pgPhaseStep_0 = _POW2_TABLE[v & _POW2T_OCTV_MASK] >> _POW2T_RS_TABLE[v >> _POW2T_OCTV_BITS];
	}
	// Inst op2 --->
	if (state & 0b0010)
	{
		v = _pgKey_1 + _pgBend + _lfoOutPM_1;
		_pgPhaseStep_1 = _POW2_TABLE[v & _POW2T_OCTV_MASK] >> _POW2T_RS_TABLE[v >> _POW2T_OCTV_BITS];
	}
	// Inst op2 --->
	if (state & 0b0100)
	{
		v = _pgKey_2 + _pgBend + _lfoOutPM_2;
		_pgPhaseStep_2 = _POW2_TABLE[v & _POW2T_OCTV_MASK] >> _POW2T_RS_TABLE[v >> _POW2T_OCTV_BITS];
	}
	// Inst op3 --->
	if (state & 0b1000)
	{
		v = _pgKey_3 + _pgBend + _lfoOutPM_3;
		_pgPhaseStep_3 = _POW2_TABLE[v & _POW2T_OCTV_MASK] >> _POW2T_RS_TABLE[v >> _POW2T_OCTV_BITS];
	}
}
void SynthesizerFM4::VoiceUnitFM4::updPartModulation(const int32& val)
{
	_lfoMod_0 = val * _lfoModSens_0;
	_lfoMod_1 = val * _lfoModSens_1;
	_lfoMod_2 = val * _lfoModSens_2;
	_lfoMod_3 = val * _lfoModSens_3;
}
void SynthesizerFM4::VoiceUnitFM4::updateLfoFrequency(const int32& val)
{
	_lfoTimerStep = val * lfoTimerCoef;
}
void SynthesizerFM4::VoiceUnitFM4::_proc0(float* const& outL, float* const& outR, const int32& end)
{
	const int32* _et_0 = _egLevelTable_0, *_et_1 = _egLevelTable_1,
	            *_et_2 = _egLevelTable_2, *_et_3 = _egLevelTable_3;
	const float* const& _log2vT = _LOG2V_TABLE;
	const int32* const& _log2vTI = _LOG2V_TABLE_I;
	const int32* const& _wt_0 = _waveTable_0, * const& _wt_1 = _waveTable_1,
	           * const& _wt_2 = _waveTable_2, * const& _wt_3 = _waveTable_3;
	int x, p, f; float n = 0.0f;
	for (int& i = startFrame; i < end; ++i)
	{
		// lfo Update --->
		if (_lfoTimer <= 0)
		{
			p = _lfoWaveTable[_lfoPhase];
			// op0
			_lfoOutPM_0 = (p * _lfoMod_0) >> 11;  // 7(mod) + 4(sens)
			x = _pgKey_0 + _pgBend + _lfoOutPM_0;
			_pgPhaseStep_0 = _POW2_TABLE[x & _POW2T_OCTV_MASK] >> _POW2T_RS_TABLE[x >> _POW2T_OCTV_BITS];
			// op1
			_lfoOutPM_1 = (p * _lfoMod_1) >> 11;
			x = _pgKey_1 + _pgBend + _lfoOutPM_1;
			_pgPhaseStep_1 = _POW2_TABLE[x & _POW2T_OCTV_MASK] >> _POW2T_RS_TABLE[x >> _POW2T_OCTV_BITS];
			// op0
			_lfoOutPM_2 = (p * _lfoMod_2) >> 11;
			x = _pgKey_2 + _pgBend + _lfoOutPM_2;
			_pgPhaseStep_2 = _POW2_TABLE[x & _POW2T_OCTV_MASK] >> _POW2T_RS_TABLE[x >> _POW2T_OCTV_BITS];
			// op1
			_lfoOutPM_3 = (p * _lfoMod_3) >> 11;
			x = _pgKey_3 + _pgBend + _lfoOutPM_3;
			_pgPhaseStep_3 = _POW2_TABLE[x & _POW2T_OCTV_MASK] >> _POW2T_RS_TABLE[x >> _POW2T_OCTV_BITS];
			// counter
			_lfoPhase = _lfoPhase + 2 & _LFO_WAVE_TABLE_MASK;
			_lfoTimer += _LFO_TIMER_INIT_VALUE;
		}
		// Operator 0 --->
		if (state & 0b0001)
		{
		// eg update --->
			if (_egCount_0 >= _egGate_0)
			{
				if (_shiftEGState(0))_et_0 = _egLevelTable_0;
				else
				{
					if (state & 0b1110){ state &= 0b11111110; _egCount_0 = _EG_MAX_CNT - 1; }
					else { killProcess(); return; }
				}
			}
			x = _et_0[_egCount_0 >> _EG_CNT_FP] + _egTLv_0;
			// pg update --->
			p = _pgPhase_0 >> _PG_PHASE_FP;
			// FeedBack -->
			f = _log2vTI[_wt_0[p] + _fdbk];
			// OP 0 out --->
			p = _wt_0[p + f & _WAVE_TABLE_MASK] + x;
			// OP 0 out --->
			n = _log2vT[p];
		// counter --->
			_egCount_0 += _egCountStep_0;
			_pgPhase_0 = _pgPhase_0 + _pgPhaseStep_0 & _PG_PHASE_MASK;
		} else n = 0.0f;
		// Operator 1 --->
		if (state & 0b0010)
		{
			// eg update --->
			if (_egCount_1 >= _egGate_1)
			{
				if (_shiftEGState(1))_et_1 = _egLevelTable_1;
				else
				{
					if (state & 0b1101) { state &= 0b11111101; _egCount_1 = _EG_MAX_CNT - 1; }
					else { killProcess(); return; }
				}
			}
			x = _et_1[_egCount_1 >> _EG_CNT_FP] + _egTLv_1;
			// pg update --->
			x = _wt_1[_pgPhase_1 >> _PG_PHASE_FP] + x;
			// OP 1 out ----->
			n += _log2vT[x];
			// counter --->
			_egCount_1 += _egCountStep_1;
			_pgPhase_1 = _pgPhase_1 + _pgPhaseStep_1 & _PG_PHASE_MASK;
		}
		// Operator 2 --->
		if (state & 0b0100)
		{
			// eg update --->
			if (_egCount_2 >= _egGate_2)
			{
				if (_shiftEGState(2))_et_2 = _egLevelTable_2;
				else
				{
					if (state & 0b1011) { state &= 0b11111011; _egCount_2 = _EG_MAX_CNT - 1; }
					else { killProcess(); return; }
				}
			}
			x = _et_2[_egCount_2 >> _EG_CNT_FP] + _egTLv_2;
			// pg update --->
			x = _wt_2[_pgPhase_2 >> _PG_PHASE_FP] + x;
			// OP 1 out ----->
			n += _log2vT[x];
			// counter --->
			_egCount_2 += _egCountStep_2;
			_pgPhase_2 = _pgPhase_2 + _pgPhaseStep_2 & _PG_PHASE_MASK;
		}
		// Operator 3 --->
		if (state & 0b1000)
		{
			// eg update --->
			if (_egCount_3 >= _egGate_3)
			{
				if (_shiftEGState(3))_et_3 = _egLevelTable_3;
				else
				{
					if (state & 0b0111) { state &= 0b11110111; _egCount_3 = _EG_MAX_CNT - 1; }
					else { killProcess(); return; }
				}
			}
			x = _et_3[_egCount_3 >> _EG_CNT_FP] + _egTLv_3;
			// pg update --->
			x = _wt_3[_pgPhase_3 >> _PG_PHASE_FP] + x;
			// OP 1 out ----->
			n += _log2vT[x];
			// counter --->
			_egCount_3 += _egCountStep_3;
			_pgPhase_3 = _pgPhase_3 + _pgPhaseStep_3 & _PG_PHASE_MASK;
		}
		// out --->
		outL[i] += n * _vOutL;
		outR[i] += n * _vOutR;
	}
}
void SynthesizerFM4::VoiceUnitFM4::_proc1(float* const& outL, float* const& outR, const int32& end)
{
	const int32* _et_0 = _egLevelTable_0, *_et_1 = _egLevelTable_1,
	            *_et_2 = _egLevelTable_2, *_et_3 = _egLevelTable_3;
	const float* const& _log2vT = _LOG2V_TABLE;
	const int32* const& _log2vTI = _LOG2V_TABLE_I;
	const int32* const& _wt_0 = _waveTable_0, * const& _wt_1 = _waveTable_1,
	           * const& _wt_2 = _waveTable_2, * const& _wt_3 = _waveTable_3;
	int x, p, f, o; float n = 0.0f;
	for (int& i = startFrame; i < end; ++i)
	{
	// lfo Update --->
		if (_lfoTimer <= 0)
		{
			p = _lfoWaveTable[_lfoPhase];
			// op0
			_lfoOutPM_0 = (p * _lfoMod_0) >> 11;  // 7(mod) + 4(sens)
			x = _pgKey_0 + _pgBend + _lfoOutPM_0;
			_pgPhaseStep_0 = _POW2_TABLE[x & _POW2T_OCTV_MASK] >> _POW2T_RS_TABLE[x >> _POW2T_OCTV_BITS];
			// op1
			_lfoOutPM_1 = (p * _lfoMod_1) >> 11;
			x = _pgKey_1 + _pgBend + _lfoOutPM_1;
			_pgPhaseStep_1 = _POW2_TABLE[x & _POW2T_OCTV_MASK] >> _POW2T_RS_TABLE[x >> _POW2T_OCTV_BITS];
			// op0
			_lfoOutPM_2 = (p * _lfoMod_2) >> 11;
			x = _pgKey_2 + _pgBend + _lfoOutPM_2;
			_pgPhaseStep_2 = _POW2_TABLE[x & _POW2T_OCTV_MASK] >> _POW2T_RS_TABLE[x >> _POW2T_OCTV_BITS];
			// op1
			_lfoOutPM_3 = (p * _lfoMod_3) >> 11;
			x = _pgKey_3 + _pgBend + _lfoOutPM_3;
			_pgPhaseStep_3 = _POW2_TABLE[x & _POW2T_OCTV_MASK] >> _POW2T_RS_TABLE[x >> _POW2T_OCTV_BITS];
			// counter
			_lfoPhase = _lfoPhase + 2 & _LFO_WAVE_TABLE_MASK;
			_lfoTimer += _LFO_TIMER_INIT_VALUE;
		}
	// Operator 0 --->
		if (state & 0b0001)
		{
			// eg update --->
			if (_egCount_0 >= _egGate_0)
			{
				if (_shiftEGState(0))_et_0 = _egLevelTable_0;
				else
				{
					if (state & 0b1110) { state &= 0b11111110; _egCount_0 = _EG_MAX_CNT - 1; }
					else { killProcess(); return; }
				}
			}
			x = _et_0[_egCount_0 >> _EG_CNT_FP] + _egTLv_0;
			// pg update --->
			p = _pgPhase_0 >> _PG_PHASE_FP;
			// FeedBack -->
			f = _log2vTI[_wt_0[p] + _fdbk];
			// OP 0 out --->
			p = _wt_0[p + f & _WAVE_TABLE_MASK] + x;
			o = _log2vTI[p];
			// counter --->	
			_egCount_0 += _egCountStep_0;
			_pgPhase_0 = _pgPhase_0 + _pgPhaseStep_0 & _PG_PHASE_MASK;
		}
		else o = 0;
	// Operator 1 --->
		if (state & 0b0010)
		{
		// eg update --->
			if (_egCount_1 >= _egGate_1)
			{
				if (_shiftEGState(1))_et_1 = _egLevelTable_1;
				else
				{
					if (state & 0b1101) { state &= 0b11111101; _egCount_1 = _EG_MAX_CNT - 1; }
					else { killProcess(); return; }
				}
			}
			x = _et_1[_egCount_1 >> _EG_CNT_FP] + _egTLv_1;
		// pg update --->
			p = (o + (_pgPhase_1 >> _PG_PHASE_FP)) & _WAVE_TABLE_MASK;
			x = _wt_1[p] + x;
		// OP 1 out ----->
			n = _log2vT[x];
		// counter --->
			_egCount_1 += _egCountStep_1;
			_pgPhase_1 = _pgPhase_1 + _pgPhaseStep_1 & _PG_PHASE_MASK;
		}
		else n = 0.0f;
	// Operator 2 --->
		if (state & 0b0100)
		{
			// eg update --->
			if (_egCount_2 >= _egGate_2)
			{
				if (_shiftEGState(2))_et_2 = _egLevelTable_2;
				else
				{
					if (state & 0b1011) { state &= 0b11111011; _egCount_2 = _EG_MAX_CNT - 1; }
					else { killProcess(); return; }
				}
			}
			x = _et_2[_egCount_2 >> _EG_CNT_FP] + _egTLv_2;
			// pg update --->
			x = _wt_2[_pgPhase_2 >> _PG_PHASE_FP] + x;
			// OP 1 out ----->
			n += _log2vT[x];
			// counter --->
			_egCount_2 += _egCountStep_2;
			_pgPhase_2 = _pgPhase_2 + _pgPhaseStep_2 & _PG_PHASE_MASK;
		}
	// Operator 3 --->
		if (state & 0b1000)
		{
			// eg update --->
			if (_egCount_3 >= _egGate_3)
			{
				if (_shiftEGState(3))_et_3 = _egLevelTable_3;
				else
				{
					if (state & 0b0111) { state &= 0b11110111; _egCount_3 = _EG_MAX_CNT - 1; }
					else { killProcess(); return; }
				}
			}
			x = _et_3[_egCount_3 >> _EG_CNT_FP] + _egTLv_3;
			// pg update --->
			x = _wt_3[_pgPhase_3 >> _PG_PHASE_FP] + x;
			// OP 1 out ----->
			n += _log2vT[x];
			// counter --->
			_egCount_3 += _egCountStep_3;
			_pgPhase_3 = _pgPhase_3 + _pgPhaseStep_3 & _PG_PHASE_MASK;
		}
	// out --->
		outL[i] += n * _vOutL;
		outR[i] += n * _vOutR;
	}
}
void SynthesizerFM4::VoiceUnitFM4::_proc2(float* const& outL, float* const& outR, const int32& end)
{
	const int32* _et_0 = _egLevelTable_0, *_et_1 = _egLevelTable_1,
	            *_et_2 = _egLevelTable_2, *_et_3 = _egLevelTable_3;
	const float* const& _log2vT = _LOG2V_TABLE;
	const int32* const& _log2vTI = _LOG2V_TABLE_I;
	const int32* const& _wt_0 = _waveTable_0, * const& _wt_1 = _waveTable_1,
	           * const& _wt_2 = _waveTable_2, * const& _wt_3 = _waveTable_3;
	int x, p, f, o; float n = 0.0f;
	for (int& i = startFrame; i < end; ++i)
	{
	// lfo Update --->
		if (_lfoTimer <= 0)
		{
			p = _lfoWaveTable[_lfoPhase];
			// op0
			_lfoOutPM_0 = (p * _lfoMod_0) >> 11;  // 7(mod) + 4(sens)
			x = _pgKey_0 + _pgBend + _lfoOutPM_0;
			_pgPhaseStep_0 = _POW2_TABLE[x & _POW2T_OCTV_MASK] >> _POW2T_RS_TABLE[x >> _POW2T_OCTV_BITS];
			// op1
			_lfoOutPM_1 = (p * _lfoMod_1) >> 11;
			x = _pgKey_1 + _pgBend + _lfoOutPM_1;
			_pgPhaseStep_1 = _POW2_TABLE[x & _POW2T_OCTV_MASK] >> _POW2T_RS_TABLE[x >> _POW2T_OCTV_BITS];
			// op0
			_lfoOutPM_2 = (p * _lfoMod_2) >> 11;
			x = _pgKey_2 + _pgBend + _lfoOutPM_2;
			_pgPhaseStep_2 = _POW2_TABLE[x & _POW2T_OCTV_MASK] >> _POW2T_RS_TABLE[x >> _POW2T_OCTV_BITS];
			// op1
			_lfoOutPM_3 = (p * _lfoMod_3) >> 11;
			x = _pgKey_3 + _pgBend + _lfoOutPM_3;
			_pgPhaseStep_3 = _POW2_TABLE[x & _POW2T_OCTV_MASK] >> _POW2T_RS_TABLE[x >> _POW2T_OCTV_BITS];
			// counter
			_lfoPhase = _lfoPhase + 2 & _LFO_WAVE_TABLE_MASK;
			_lfoTimer += _LFO_TIMER_INIT_VALUE;
		}
	// Operator 0 --->
		if (state & 0b0001)
		{
			// eg update --->
			if (_egCount_0 >= _egGate_0)
			{
				if (_shiftEGState(0))_et_0 = _egLevelTable_0;
				else
				{
					if (state & 0b1110) { state &= 0b11111110; _egCount_0 = _EG_MAX_CNT - 1; }
					else { killProcess(); return; }
				}
			}
			x = _et_0[_egCount_0 >> _EG_CNT_FP] + _egTLv_0;
			// pg update --->
			p = _pgPhase_0 >> _PG_PHASE_FP;
			// FeedBack -->
			f = _log2vTI[_wt_0[p] + _fdbk];
			// OP 0 out --->
			p = _wt_0[p + f & _WAVE_TABLE_MASK] + x;
			o = _log2vTI[p];
			// counter --->	
			_egCount_0 += _egCountStep_0;
			_pgPhase_0 = _pgPhase_0 + _pgPhaseStep_0 & _PG_PHASE_MASK;
		}
		else o = 0;
	// Operator 1 --->
		if (state & 0b0010)
		{
			// eg update --->
			if (_egCount_1 >= _egGate_1)
			{
				if (_shiftEGState(1))_et_1 = _egLevelTable_1;
				else
				{
					if (state & 0b1101) { state &= 0b11111101; _egCount_1 = _EG_MAX_CNT - 1; }
					else { killProcess(); return; }
				}
			}
			x = _et_1[_egCount_1 >> _EG_CNT_FP] + _egTLv_1;
			// pg update --->
			p = (o + (_pgPhase_1 >> _PG_PHASE_FP)) & _WAVE_TABLE_MASK;
			x = _wt_1[p] + x;
			// OP 1 out ----->
			n = _log2vT[x];
			// counter --->
			_egCount_1 += _egCountStep_1;
			_pgPhase_1 = _pgPhase_1 + _pgPhaseStep_1 & _PG_PHASE_MASK;
		}
		else n = 0.0f;
	// Operator 2 --->
		if (state & 0b0100)
		{
			// eg update --->
			if (_egCount_2 >= _egGate_2)
			{
				if (_shiftEGState(2))_et_2 = _egLevelTable_2;
				else
				{
					if (state & 0b1011) { state &= 0b11111011; _egCount_2 = _EG_MAX_CNT - 1; }
					else { killProcess(); return; }
				}
			}
			x = _et_2[_egCount_2 >> _EG_CNT_FP] + _egTLv_2;
			// pg update --->
			p = (o + (_pgPhase_2 >> _PG_PHASE_FP)) & _WAVE_TABLE_MASK;
			x = _wt_2[p] + x;
			// OP 1 out ----->
			n += _log2vT[x];
			// counter --->
			_egCount_2 += _egCountStep_2;
			_pgPhase_2 = _pgPhase_2 + _pgPhaseStep_2 & _PG_PHASE_MASK;
		}
	// Operator 3 --->
		if (state & 0b1000)
		{
			// eg update --->
			if (_egCount_3 >= _egGate_3)
			{
				if (_shiftEGState(3))_et_3 = _egLevelTable_3;
				else
				{
					if (state & 0b0111) { state &= 0b11110111; _egCount_3 = _EG_MAX_CNT - 1; }
					else { killProcess(); return; }
				}
			}
			x = _et_3[_egCount_3 >> _EG_CNT_FP] + _egTLv_3;
			// pg update --->
			p = (o + (_pgPhase_3 >> _PG_PHASE_FP)) & _WAVE_TABLE_MASK;
			x = _wt_3[p] + x;
			// OP 1 out ----->
			n += _log2vT[x];
			// counter --->
			_egCount_3 += _egCountStep_3;
			_pgPhase_3 = _pgPhase_3 + _pgPhaseStep_3 & _PG_PHASE_MASK;
		}
	// out --->
		outL[i] += n * _vOutL;
		outR[i] += n * _vOutR;
	// counter --->
		_lfoTimer -= _lfoTimerStep;
	}
}
void SynthesizerFM4::VoiceUnitFM4::_proc3(float* const& outL, float* const& outR, const int32& end)
{
	const int32* _et_0 = _egLevelTable_0, *_et_1 = _egLevelTable_1,
	            *_et_2 = _egLevelTable_2, *_et_3 = _egLevelTable_3;
	const float* const& _log2vT = _LOG2V_TABLE;
	const int32* const& _log2vTI = _LOG2V_TABLE_I;
	const int32* const& _wt_0 = _waveTable_0, * const& _wt_1 = _waveTable_1,
	           * const& _wt_2 = _waveTable_2, * const& _wt_3 = _waveTable_3;
	int x, p, f, o; float n = 0.0f;
	for (int& i = startFrame; i < end; ++i)
	{
	// lfo Update --->
		if (_lfoTimer <= 0)
		{
			p = _lfoWaveTable[_lfoPhase];
			// op0
			_lfoOutPM_0 = (p * _lfoMod_0) >> 11;  // 7(mod) + 4(sens)
			x = _pgKey_0 + _pgBend + _lfoOutPM_0;
			_pgPhaseStep_0 = _POW2_TABLE[x & _POW2T_OCTV_MASK] >> _POW2T_RS_TABLE[x >> _POW2T_OCTV_BITS];
			// op1
			_lfoOutPM_1 = (p * _lfoMod_1) >> 11;
			x = _pgKey_1 + _pgBend + _lfoOutPM_1;
			_pgPhaseStep_1 = _POW2_TABLE[x & _POW2T_OCTV_MASK] >> _POW2T_RS_TABLE[x >> _POW2T_OCTV_BITS];
			// op0
			_lfoOutPM_2 = (p * _lfoMod_2) >> 11;
			x = _pgKey_2 + _pgBend + _lfoOutPM_2;
			_pgPhaseStep_2 = _POW2_TABLE[x & _POW2T_OCTV_MASK] >> _POW2T_RS_TABLE[x >> _POW2T_OCTV_BITS];
			// op1
			_lfoOutPM_3 = (p * _lfoMod_3) >> 11;
			x = _pgKey_3 + _pgBend + _lfoOutPM_3;
			_pgPhaseStep_3 = _POW2_TABLE[x & _POW2T_OCTV_MASK] >> _POW2T_RS_TABLE[x >> _POW2T_OCTV_BITS];
			// counter
			_lfoPhase = _lfoPhase + 2 & _LFO_WAVE_TABLE_MASK;
			_lfoTimer += _LFO_TIMER_INIT_VALUE;
		}
	// Operator 0 --->
		if (state & 0b0001)
		{
			// eg update --->
			if (_egCount_0 >= _egGate_0)
			{
				if (_shiftEGState(0))_et_0 = _egLevelTable_0;
				else
				{
					if (state & 0b1110) { state &= 0b11111110; _egCount_0 = _EG_MAX_CNT - 1; }
					else { killProcess(); return; }
				}
			}
			x = _et_0[_egCount_0 >> _EG_CNT_FP] + _egTLv_0;
			// pg update --->
			p = _pgPhase_0 >> _PG_PHASE_FP;
			// FeedBack -->
			f = _log2vTI[_wt_0[p] + _fdbk];
			// OP 0 out --->
			x = _wt_0[p + f & _WAVE_TABLE_MASK] + x;
			o = _log2vTI[x];
			// counter --->	
			_egCount_0 += _egCountStep_0;
			_pgPhase_0 = _pgPhase_0 + _pgPhaseStep_0 & _PG_PHASE_MASK;
		}
		else o = 0;
	// Operator 1 --->
		if (state & 0b0010)
		{
			// eg update --->
			if (_egCount_1 >= _egGate_1)
			{
				if (_shiftEGState(1))_et_1 = _egLevelTable_1;
				else
				{
					if (state & 0b1101) { state &= 0b11111101; _egCount_1 = _EG_MAX_CNT - 1; }
					else { killProcess(); return; }
				}
			}
			x = _et_1[_egCount_1 >> _EG_CNT_FP] + _egTLv_1;
			// pg update --->
			p = (o + (_pgPhase_1 >> _PG_PHASE_FP)) & _WAVE_TABLE_MASK;
			x = _wt_1[p] + x;
			// OP 1 out ----->
			n = _log2vT[x];
			// counter --->
			_egCount_1 += _egCountStep_1;
			_pgPhase_1 = _pgPhase_1 + _pgPhaseStep_1 & _PG_PHASE_MASK;
		}
		else n = 0.0f;
	// Operator 2 --->
		if (state & 0b0100)
		{
			// eg update --->
			if (_egCount_2 >= _egGate_2)
			{
				if (_shiftEGState(2))_et_2 = _egLevelTable_2;
				else
				{
					if (state & 0b1011) { state &= 0b11111011; _egCount_2 = _EG_MAX_CNT - 1; }
					else { killProcess(); return; }
				}
			}
			x = _et_2[_egCount_2 >> _EG_CNT_FP] + _egTLv_2;
			// pg update --->
			x = _wt_2[_pgPhase_2 >> _PG_PHASE_FP] + x;
			// OP 0 out --->
			o = _log2vTI[x];
			// counter --->	
			_egCount_2 += _egCountStep_2;
			_pgPhase_2 = _pgPhase_2 + _pgPhaseStep_2 & _PG_PHASE_MASK;
		}
		else o = 0;
	// Operator 3 --->
		if (state & 0b1000)
		{
			// eg update --->
			if (_egCount_3 >= _egGate_3)
			{
				if (_shiftEGState(3))_et_3 = _egLevelTable_3;
				else
				{
					if (state & 0b0111) { state &= 0b11110111; _egCount_3 = _EG_MAX_CNT - 1; }
					else { killProcess(); return; }
				}
			}
			x = _et_3[_egCount_3 >> _EG_CNT_FP] + _egTLv_3;
			// pg update --->
			p = (o + (_pgPhase_3 >> _PG_PHASE_FP)) & _WAVE_TABLE_MASK;
			x = _wt_3[p] + x;
			// OP 1 out ----->
			n += _log2vT[x];
			// counter --->
			_egCount_3 += _egCountStep_3;
			_pgPhase_3 = _pgPhase_3 + _pgPhaseStep_3 & _PG_PHASE_MASK;
		}
		// out --->
		outL[i] += n * _vOutL;
		outR[i] += n * _vOutR;
	// counter --->
		_lfoTimer -= _lfoTimerStep;
	}
}
void SynthesizerFM4::VoiceUnitFM4::_proc4(float* const& outL, float* const& outR, const int32& end)
{
	const int32* _et_0 = _egLevelTable_0, *_et_1 = _egLevelTable_1,
	            *_et_2 = _egLevelTable_2, *_et_3 = _egLevelTable_3;
	const float* const& _log2vT = _LOG2V_TABLE;
	const int32* const& _log2vTI = _LOG2V_TABLE_I;
	const int32* const& _wt_0 = _waveTable_0, * const& _wt_1 = _waveTable_1,
	           * const& _wt_2 = _waveTable_2, * const& _wt_3 = _waveTable_3;
	int x, p, f, o; float n = 0.0f;
	for (int& i = startFrame; i < end; ++i)
	{
	// lfo Update --->
		if (_lfoTimer <= 0)
		{
			p = _lfoWaveTable[_lfoPhase];
			// op0
			_lfoOutPM_0 = (p * _lfoMod_0) >> 11;  // 7(mod) + 4(sens)
			x = _pgKey_0 + _pgBend + _lfoOutPM_0;
			_pgPhaseStep_0 = _POW2_TABLE[x & _POW2T_OCTV_MASK] >> _POW2T_RS_TABLE[x >> _POW2T_OCTV_BITS];
			// op1
			_lfoOutPM_1 = (p * _lfoMod_1) >> 11;
			x = _pgKey_1 + _pgBend + _lfoOutPM_1;
			_pgPhaseStep_1 = _POW2_TABLE[x & _POW2T_OCTV_MASK] >> _POW2T_RS_TABLE[x >> _POW2T_OCTV_BITS];
			// op0
			_lfoOutPM_2 = (p * _lfoMod_2) >> 11;
			x = _pgKey_2 + _pgBend + _lfoOutPM_2;
			_pgPhaseStep_2 = _POW2_TABLE[x & _POW2T_OCTV_MASK] >> _POW2T_RS_TABLE[x >> _POW2T_OCTV_BITS];
			// op1
			_lfoOutPM_3 = (p * _lfoMod_3) >> 11;
			x = _pgKey_3 + _pgBend + _lfoOutPM_3;
			_pgPhaseStep_3 = _POW2_TABLE[x & _POW2T_OCTV_MASK] >> _POW2T_RS_TABLE[x >> _POW2T_OCTV_BITS];
			// counter
			_lfoPhase = _lfoPhase + 2 & _LFO_WAVE_TABLE_MASK;
			_lfoTimer += _LFO_TIMER_INIT_VALUE;
		}
	// Operator 0 --->
		if (state & 0b0001)
		{
			// eg update --->
			if (_egCount_0 >= _egGate_0)
			{
				if (_shiftEGState(0))_et_0 = _egLevelTable_0;
				else
				{
					if (state & 0b1110) { state &= 0b11111110; _egCount_0 = _EG_MAX_CNT - 1; }
					else { killProcess(); return; }
				}
			}
			x = _et_0[_egCount_0 >> _EG_CNT_FP] + _egTLv_0;
			// pg update --->
			p = _pgPhase_0 >> _PG_PHASE_FP;
			// FeedBack -->
			f = _log2vTI[_wt_0[p] + _fdbk];
			// OP 0 out --->
			x = _wt_0[p + f & _WAVE_TABLE_MASK] + x;
			o = _log2vTI[x];
			// counter --->	
			_egCount_0 += _egCountStep_0;
			_pgPhase_0 = _pgPhase_0 + _pgPhaseStep_0 & _PG_PHASE_MASK;
		}
		else o = 0;
	// Operator 1 --->
		if (state & 0b0010)
		{
		// eg update --->
			if (_egCount_1 >= _egGate_1)
			{
				if (_shiftEGState(1))_et_1 = _egLevelTable_1;
				else
				{
					if (state & 0b1101) { state &= 0b11111101; _egCount_1 = _EG_MAX_CNT - 1; }
					else { killProcess(); return; }
				}
			}
			x = _et_1[_egCount_1 >> _EG_CNT_FP] + _egTLv_1;
		// pg update --->
			p = (o + (_pgPhase_1 >> _PG_PHASE_FP)) & _WAVE_TABLE_MASK;
			x = _wt_1[p] + x;
		// OP 1 out ----->
			o = _log2vTI[x];
		// counter --->
			_egCount_1 += _egCountStep_1;
			_pgPhase_1 = _pgPhase_1 + _pgPhaseStep_1 & _PG_PHASE_MASK;
		}
		else o = 0;
	// Operator 2 --->
		if (state & 0b0100)
		{
			// eg update --->
			if (_egCount_2 >= _egGate_2)
			{
				if (_shiftEGState(2))_et_2 = _egLevelTable_2;
				else
				{
					if (state & 0b1011) { state &= 0b11111011; _egCount_2 = _EG_MAX_CNT - 1; }
					else { killProcess(); return; }
				}
			}
			x = _et_2[_egCount_2 >> _EG_CNT_FP] + _egTLv_2;
			// pg update --->
			x = _wt_2[_pgPhase_2 >> _PG_PHASE_FP] + x;
			// OP 1 out ----->
			o += _log2vTI[x];
			// counter --->
			_egCount_2 += _egCountStep_2;
			_pgPhase_2 = _pgPhase_2 + _pgPhaseStep_2 & _PG_PHASE_MASK;
		}
	// Operator 3 --->
		if (state & 0b1000)
		{
			// eg update --->
			if (_egCount_3 >= _egGate_3)
			{
				if (_shiftEGState(3))_et_3 = _egLevelTable_3;
				else { killProcess(); return; }
			}
			x = _et_3[_egCount_3 >> _EG_CNT_FP] + _egTLv_3;
			// pg update --->
			p = (o + (_pgPhase_3 >> _PG_PHASE_FP)) & _WAVE_TABLE_MASK;
			x = _wt_3[p] + x;
			// OP 1 out ----->
			n = _log2vT[x];
			// counter --->
			_egCount_3 += _egCountStep_3;
			_pgPhase_3 = _pgPhase_3 + _pgPhaseStep_3 & _PG_PHASE_MASK;
		}
	// out --->
		outL[i] += n * _vOutL;
		outR[i] += n * _vOutR;
	}
}
void SynthesizerFM4::VoiceUnitFM4::_proc5(float* const& outL, float* const& outR, const int32& end)
{
	const int32* _et_0 = _egLevelTable_0, *_et_1 = _egLevelTable_1,
	            *_et_2 = _egLevelTable_2, *_et_3 = _egLevelTable_3;
	const float* const& _log2vT = _LOG2V_TABLE;
	const int32* const& _log2vTI = _LOG2V_TABLE_I;
	const int32* const& _wt_0 = _waveTable_0, * const& _wt_1 = _waveTable_1,
	           * const& _wt_2 = _waveTable_2, * const& _wt_3 = _waveTable_3;
	int x, p, f, o; float n = 0.0f;
	for (int& i = startFrame; i < end; ++i)
	{
	// lfo Update --->
		if (_lfoTimer <= 0)
		{
			p = _lfoWaveTable[_lfoPhase];
			// op0
			_lfoOutPM_0 = (p * _lfoMod_0) >> 11;  // 7(mod) + 4(sens)
			x = _pgKey_0 + _pgBend + _lfoOutPM_0;
			_pgPhaseStep_0 = _POW2_TABLE[x & _POW2T_OCTV_MASK] >> _POW2T_RS_TABLE[x >> _POW2T_OCTV_BITS];
			// op1
			_lfoOutPM_1 = (p * _lfoMod_1) >> 11;
			x = _pgKey_1 + _pgBend + _lfoOutPM_1;
			_pgPhaseStep_1 = _POW2_TABLE[x & _POW2T_OCTV_MASK] >> _POW2T_RS_TABLE[x >> _POW2T_OCTV_BITS];
			// op0
			_lfoOutPM_2 = (p * _lfoMod_2) >> 11;
			x = _pgKey_2 + _pgBend + _lfoOutPM_2;
			_pgPhaseStep_2 = _POW2_TABLE[x & _POW2T_OCTV_MASK] >> _POW2T_RS_TABLE[x >> _POW2T_OCTV_BITS];
			// op1
			_lfoOutPM_3 = (p * _lfoMod_3) >> 11;
			x = _pgKey_3 + _pgBend + _lfoOutPM_3;
			_pgPhaseStep_3 = _POW2_TABLE[x & _POW2T_OCTV_MASK] >> _POW2T_RS_TABLE[x >> _POW2T_OCTV_BITS];
			// counter
			_lfoPhase = _lfoPhase + 2 & _LFO_WAVE_TABLE_MASK;
			_lfoTimer += _LFO_TIMER_INIT_VALUE;
		}
	// Operator 1 --->
		if (state & 0b0010)
		{
			// eg update --->
			if (_egCount_1 >= _egGate_1)
			{
				if (_shiftEGState(1))_et_1 = _egLevelTable_1;
				else
				{
					if (state & 0b1101) { state &= 0b11111101; _egCount_1 = _EG_MAX_CNT - 1; }
					else { killProcess(); return; }
				}
			}
			x = _et_1[_egCount_1 >> _EG_CNT_FP] + _egTLv_1;
			// pg update --->
			p = _wt_1[_pgPhase_1 >> _PG_PHASE_FP] + x;
			// OP 1 out ----->
			o = _log2vTI[p];
			// counter --->
			_egCount_1 += _egCountStep_1;
			_pgPhase_1 = _pgPhase_1 + _pgPhaseStep_1 & _PG_PHASE_MASK;
		}
		else o = 0;
	// Operator 2 --->
		if (state & 0b0100)
		{
			// eg update --->
			if (_egCount_2 >= _egGate_2)
			{
				if (_shiftEGState(2))_et_2 = _egLevelTable_2;
				else
				{
					if (state & 0b1011) { state &= 0b11111011; _egCount_2 = _EG_MAX_CNT - 1; }
					else { killProcess(); return; }
				}
			}
			x = _et_2[_egCount_2 >> _EG_CNT_FP] + _egTLv_2;
			// pg update --->
			p = (o + (_pgPhase_2 >> _PG_PHASE_FP)) & _WAVE_TABLE_MASK;
			x = _wt_2[p] + x;
			// OP 1 out ----->
			o = _log2vTI[x];
			// counter --->
			_egCount_2 += _egCountStep_2;
			_pgPhase_2 = _pgPhase_2 + _pgPhaseStep_2 & _PG_PHASE_MASK;
		}
		else o = 0;
	// Operator 0 --->
		if (state & 0b0001)
		{
			// eg update --->
			if (_egCount_0 >= _egGate_0)
			{
				if (_shiftEGState(0))_et_0 = _egLevelTable_0;
				else
				{
					if (state & 0b1110) { state &= 0b11111110; _egCount_0 = _EG_MAX_CNT - 1; }
					else { killProcess(); return; }
				}
			}
			x = _et_0[_egCount_0 >> _EG_CNT_FP] + _egTLv_0;
			// pg update --->
			p = _pgPhase_0 >> _PG_PHASE_FP;
			// FeedBack -->
			f = _log2vTI[_wt_0[p] + _fdbk];
			// OP 0 out --->
			x = _wt_0[p + f & _WAVE_TABLE_MASK] + x;
			o += _log2vTI[x];
			// counter --->	
			_egCount_0 += _egCountStep_0;
			_pgPhase_0 = _pgPhase_0 + _pgPhaseStep_0 & _PG_PHASE_MASK;
		}
	// Operator 3 --->
		if (state & 0b1000)
		{
			// eg update --->
			if (_egCount_3 >= _egGate_3)
			{
				if (_shiftEGState(3))_et_3 = _egLevelTable_3;
				else { killProcess(); return; }
			}
			x = _et_3[_egCount_3 >> _EG_CNT_FP] + _egTLv_3;
			// pg update --->
			p = (o + (_pgPhase_3 >> _PG_PHASE_FP)) & _WAVE_TABLE_MASK;
			x = _wt_3[p] + x;
			// OP 1 out ----->
			n = _log2vT[x];
			// counter --->
			_egCount_3 += _egCountStep_3;
			_pgPhase_3 = _pgPhase_3 + _pgPhaseStep_3 & _PG_PHASE_MASK;
		}
		// out --->
		outL[i] += n * _vOutL;
		outR[i] += n * _vOutR;
	}
}
void SynthesizerFM4::VoiceUnitFM4::_proc6(float* const& outL, float* const& outR, const int32& end)
{
	const int32* _et_0 = _egLevelTable_0, *_et_1 = _egLevelTable_1,
	            *_et_2 = _egLevelTable_2, *_et_3 = _egLevelTable_3;
	const float* const& _log2vT = _LOG2V_TABLE;
	const int32* const& _log2vTI = _LOG2V_TABLE_I;
	const int32* const& _wt_0 = _waveTable_0, * const& _wt_1 = _waveTable_1,
	           * const& _wt_2 = _waveTable_2, * const& _wt_3 = _waveTable_3;
	int x, p, f, o; float n = 0.0f;
	for (int& i = startFrame; i < end; ++i)
	{
	// lfo Update --->
		if (_lfoTimer <= 0)
		{
			p = _lfoWaveTable[_lfoPhase];
			// op0
			_lfoOutPM_0 = (p * _lfoMod_0) >> 11;  // 7(mod) + 4(sens)
			x = _pgKey_0 + _pgBend + _lfoOutPM_0;
			_pgPhaseStep_0 = _POW2_TABLE[x & _POW2T_OCTV_MASK] >> _POW2T_RS_TABLE[x >> _POW2T_OCTV_BITS];
			// op1
			_lfoOutPM_1 = (p * _lfoMod_1) >> 11;
			x = _pgKey_1 + _pgBend + _lfoOutPM_1;
			_pgPhaseStep_1 = _POW2_TABLE[x & _POW2T_OCTV_MASK] >> _POW2T_RS_TABLE[x >> _POW2T_OCTV_BITS];
			// op0
			_lfoOutPM_2 = (p * _lfoMod_2) >> 11;
			x = _pgKey_2 + _pgBend + _lfoOutPM_2;
			_pgPhaseStep_2 = _POW2_TABLE[x & _POW2T_OCTV_MASK] >> _POW2T_RS_TABLE[x >> _POW2T_OCTV_BITS];
			// op1
			_lfoOutPM_3 = (p * _lfoMod_3) >> 11;
			x = _pgKey_3 + _pgBend + _lfoOutPM_3;
			_pgPhaseStep_3 = _POW2_TABLE[x & _POW2T_OCTV_MASK] >> _POW2T_RS_TABLE[x >> _POW2T_OCTV_BITS];
			// counter
			_lfoPhase = _lfoPhase + 2 & _LFO_WAVE_TABLE_MASK;
			_lfoTimer += _LFO_TIMER_INIT_VALUE;
		}
	// Operator 0 --->
		if (state & 0b0001)
		{
			// eg update --->
			if (_egCount_0 >= _egGate_0)
			{
				if (_shiftEGState(0))_et_0 = _egLevelTable_0;
				else
				{
					if (state & 0b1110) { state &= 0b11111110; _egCount_0 = _EG_MAX_CNT - 1; }
					else { killProcess(); return; }
				}
			}
			x = _et_0[_egCount_0 >> _EG_CNT_FP] + _egTLv_0;
			// pg update --->
			p = _pgPhase_0 >> _PG_PHASE_FP;
			// FeedBack -->
			f = _log2vTI[_wt_0[p] + _fdbk];
			// OP 0 out --->
			x = _wt_0[p + f & _WAVE_TABLE_MASK] + x;
			o = _log2vTI[x];
			// counter --->	
			_egCount_0 += _egCountStep_0;
			_pgPhase_0 = _pgPhase_0 + _pgPhaseStep_0 & _PG_PHASE_MASK;
		}
		else o = 0;
	// Operator 1 --->
		if (state & 0b0010)
		{
		// eg update --->
			if (_egCount_1 >= _egGate_1)
			{
				if (_shiftEGState(1))_et_1 = _egLevelTable_1;
				else
				{
					if (state & 0b1101) { state &= 0b11111101; _egCount_1 = _EG_MAX_CNT - 1; }
					else { killProcess(); return; }
				}
			}
			x = _et_1[_egCount_1 >> _EG_CNT_FP] + _egTLv_1;
		// pg update --->
			x = _wt_1[_pgPhase_1 >> _PG_PHASE_FP] + x;
		// OP 1 out ----->
			o += _log2vTI[x];
		// counter --->
			_egCount_1 += _egCountStep_1;
			_pgPhase_1 = _pgPhase_1 + _pgPhaseStep_1 & _PG_PHASE_MASK;
		}
	// Operator 2 --->
		if (state & 0b0100)
		{
			// eg update --->
			if (_egCount_2 >= _egGate_2)
			{
				if (_shiftEGState(2))_et_2 = _egLevelTable_2;
				else
				{
					if (state & 0b1011) { state &= 0b11111011; _egCount_2 = _EG_MAX_CNT - 1; }
					else { killProcess(); return; }
				}
			}
			x = _et_2[_egCount_2 >> _EG_CNT_FP] + _egTLv_2;
			// pg update --->
			p = (o + (_pgPhase_2 >> _PG_PHASE_FP)) & _WAVE_TABLE_MASK;
			x = _wt_2[p] + x;
			// OP 1 out ----->
			o = _log2vTI[x];
			// counter --->
			_egCount_2 += _egCountStep_2;
			_pgPhase_2 = _pgPhase_2 + _pgPhaseStep_2 & _PG_PHASE_MASK;
		}
		else o = 0;
	// Operator 3 --->
		if (state & 0b1000)
		{
			// eg update --->
			if (_egCount_3 >= _egGate_3)
			{
				if (_shiftEGState(3))_et_3 = _egLevelTable_3;
				else { killProcess(); return; }
			}
			x = _et_3[_egCount_3 >> _EG_CNT_FP] + _egTLv_3;
			// pg update --->
			p = (o + (_pgPhase_3 >> _PG_PHASE_FP)) & _WAVE_TABLE_MASK;
			x = _wt_3[p] + x;
			// OP 1 out ----->
			n = _log2vT[x];
			// counter --->
			_egCount_3 += _egCountStep_3;
			_pgPhase_3 = _pgPhase_3 + _pgPhaseStep_3 & _PG_PHASE_MASK;
		}
	// out --->
		outL[i] += n * _vOutL;
		outR[i] += n * _vOutR;
	// counter --->
		_lfoTimer -= _lfoTimerStep;
	}
}
void SynthesizerFM4::VoiceUnitFM4::_proc7(float* const& outL, float* const& outR, const int32& end)
{
	const int32* _et_0 = _egLevelTable_0, *_et_1 = _egLevelTable_1,
	            *_et_2 = _egLevelTable_2, *_et_3 = _egLevelTable_3;
	const float* const& _log2vT = _LOG2V_TABLE;
	const int32* const& _log2vTI = _LOG2V_TABLE_I;
	const int32* const& _wt_0 = _waveTable_0, * const& _wt_1 = _waveTable_1,
	           * const& _wt_2 = _waveTable_2, * const& _wt_3 = _waveTable_3;
	int x, p, f, o; float n = 0.0f;
	for (int& i = startFrame; i < end; ++i)
	{
	// lfo Update --->
		if (_lfoTimer <= 0)
		{
			p = _lfoWaveTable[_lfoPhase];
			// op0
			_lfoOutPM_0 = (p * _lfoMod_0) >> 11;  // 7(mod) + 4(sens)
			x = _pgKey_0 + _pgBend + _lfoOutPM_0;
			_pgPhaseStep_0 = _POW2_TABLE[x & _POW2T_OCTV_MASK] >> _POW2T_RS_TABLE[x >> _POW2T_OCTV_BITS];
			// op1
			_lfoOutPM_1 = (p * _lfoMod_1) >> 11;
			x = _pgKey_1 + _pgBend + _lfoOutPM_1;
			_pgPhaseStep_1 = _POW2_TABLE[x & _POW2T_OCTV_MASK] >> _POW2T_RS_TABLE[x >> _POW2T_OCTV_BITS];
			// op0
			_lfoOutPM_2 = (p * _lfoMod_2) >> 11;
			x = _pgKey_2 + _pgBend + _lfoOutPM_2;
			_pgPhaseStep_2 = _POW2_TABLE[x & _POW2T_OCTV_MASK] >> _POW2T_RS_TABLE[x >> _POW2T_OCTV_BITS];
			// op1
			_lfoOutPM_3 = (p * _lfoMod_3) >> 11;
			x = _pgKey_3 + _pgBend + _lfoOutPM_3;
			_pgPhaseStep_3 = _POW2_TABLE[x & _POW2T_OCTV_MASK] >> _POW2T_RS_TABLE[x >> _POW2T_OCTV_BITS];
			// counter
			_lfoPhase = _lfoPhase + 2 & _LFO_WAVE_TABLE_MASK;
			_lfoTimer += _LFO_TIMER_INIT_VALUE;
		}
	// Operator 0 --->
		if (state & 0b0001)
		{
			// eg update --->
			if (_egCount_0 >= _egGate_0)
			{
				if (_shiftEGState(0))_et_0 = _egLevelTable_0;
				else
				{
					if (state & 0b1110) { state &= 0b11111110; _egCount_0 = _EG_MAX_CNT - 1; }
					else { killProcess(); return; }
				}
			}
			x = _et_0[_egCount_0 >> _EG_CNT_FP] + _egTLv_0;
			// pg update --->
			p = _pgPhase_0 >> _PG_PHASE_FP;
			// FeedBack -->
			f = _log2vTI[_wt_0[p] + _fdbk];
			// OP 0 out --->
			p = _wt_0[p + f & _WAVE_TABLE_MASK] + x;
			o = _log2vTI[p];
			// counter --->	
			_egCount_0 += _egCountStep_0;
			_pgPhase_0 = _pgPhase_0 + _pgPhaseStep_0 & _PG_PHASE_MASK;
		}
		else o = 0;
	// Operator 1 --->
		if (state & 0b0010)
		{
		// eg update --->
			if (_egCount_1 >= _egGate_1)
			{
				if (_shiftEGState(1))_et_1 = _egLevelTable_1;
				else
				{
					if (state & 0b1101) { state &= 0b11111101; _egCount_1 = _EG_MAX_CNT - 1; }
					else { killProcess(); return; }
				}
			}
			x = _et_1[_egCount_1 >> _EG_CNT_FP] + _egTLv_1;
		// pg update --->
			p = (o + (_pgPhase_1 >> _PG_PHASE_FP)) & _WAVE_TABLE_MASK;
			p = _wt_1[p] + x;
		// OP 1 out ----->
			o = _log2vTI[p];
		// counter --->
			_egCount_1 += _egCountStep_1;
			_pgPhase_1 = _pgPhase_1 + _pgPhaseStep_1 & _PG_PHASE_MASK;
		}
		else o = 0;
	// Operator 2 --->
		if (state & 0b0100)
		{
			// eg update --->
			if (_egCount_2 >= _egGate_2)
			{
				if (_shiftEGState(2))_et_2 = _egLevelTable_2;
				else
				{
					if (state & 0b1011) { state &= 0b11111011; _egCount_2 = _EG_MAX_CNT - 1; }
					else { killProcess(); return; }
				}
			}
			x = _et_2[_egCount_2 >> _EG_CNT_FP] + _egTLv_2;
			// pg update --->
			p = (o + (_pgPhase_2 >> _PG_PHASE_FP)) & _WAVE_TABLE_MASK;
			p = _wt_2[p] + x;
			// OP 1 out ----->
			o = _log2vTI[p];
			// counter --->
			_egCount_2 += _egCountStep_2;
			_pgPhase_2 = _pgPhase_2 + _pgPhaseStep_2 & _PG_PHASE_MASK;
		}
		else o = 0;
	// Operator 3 --->
		if (state & 0b1000)
		{
			// eg update --->
			if (_egCount_3 >= _egGate_3)
			{
				if (_shiftEGState(3))_et_3 = _egLevelTable_3;
				else { killProcess(); return; }
			}
			x = _et_3[_egCount_3 >> _EG_CNT_FP] + _egTLv_3;
			// pg update --->
			p = (o + (_pgPhase_3 >> _PG_PHASE_FP)) & _WAVE_TABLE_MASK;
			p = _wt_3[p] + x;
			// OP 1 out ----->
			n = _log2vT[p];
			// counter --->
			_egCount_3 += _egCountStep_3;
			_pgPhase_3 = _pgPhase_3 + _pgPhaseStep_3 & _PG_PHASE_MASK;
		}
		//else n = 0.0f;
	// out --->
		outL[i] += n * _vOutL;
		outR[i] += n * _vOutR;
	// counter --->
		_lfoTimer -= _lfoTimerStep;
	}
}

void SynthesizerFM4::VoiceUnitFM4::_proc8(float* const& outL, float* const& outR, const int32& end)
{
	const int32* _et_0 = _egLevelTable_0, *_et_1 = _egLevelTable_1,
	            *_et_2 = _egLevelTable_2, *_et_3 = _egLevelTable_3;
	const float* const& _log2vT = _LOG2V_TABLE;
	const int32* const& _log2vTI = _LOG2V_TABLE_I;
	const int32* const& _wt_0 = _waveTable_0, * const& _wt_1 = _waveTable_1,
	           * const& _wt_2 = _waveTable_2, * const& _wt_3 = _waveTable_3;
	int x, p, f; float n = 0.0f;
	for (int& i = startFrame; i < end; ++i)
	{
	// lfo Update --->
		if (_lfoTimer <= 0)
		{
			p = _lfoWaveTable[_lfoPhase];
			x = (p * _lfoMod_0) >> 11;
			_lfoOutAM_0 = x & 0x7ffffffe;
			x = (p * _lfoMod_1) >> 11;
			_lfoOutAM_1 = x & 0x7ffffffe;
			x = (p * _lfoMod_2) >> 11;
			_lfoOutAM_2 = x & 0x7ffffffe;
			x = (p * _lfoMod_3) >> 11;
			_lfoOutAM_3 = x & 0x7ffffffe;
			_lfoPhase = _lfoPhase + 2 & _LFO_WAVE_TABLE_MASK;
			_lfoTimer += _LFO_TIMER_INIT_VALUE;
		}
	// Operator 0 --->
		if (state & 0b0001)
		{
		// eg update --->
			if (_egCount_0 >= _egGate_0)
			{
				if (_shiftEGState(0))_et_0 = _egLevelTable_0;
				else
				{
					if (state & 0b1110){ state &= 0b11111110; _egCount_0 = _EG_MAX_CNT - 1; }
					else { killProcess(); return; }
				}
			}
			x = _et_0[_egCount_0 >> _EG_CNT_FP] + _egTLv_0 + _lfoOutAM_0;
			// pg update --->
			p = _pgPhase_0 >> _PG_PHASE_FP;
			// FeedBack -->
			f = _log2vTI[_wt_0[p] + _fdbk];
			// OP 0 out --->
			p = _wt_0[p + f & _WAVE_TABLE_MASK] + x;
			// OP 0 out --->
			n = _log2vT[p];
		// counter --->
			_egCount_0 += _egCountStep_0;
			_pgPhase_0 = _pgPhase_0 + _pgPhaseStep_0 & _PG_PHASE_MASK;
		} else n = 0.0f;
	// Operator 1 --->
		if (state & 0b0010)
		{
			// eg update --->
			if (_egCount_1 >= _egGate_1)
			{
				if (_shiftEGState(1))_et_1 = _egLevelTable_1;
				else
				{
					if (state & 0b1101) { state &= 0b11111101; _egCount_1 = _EG_MAX_CNT - 1; }
					else { killProcess(); return; }
				}
			}
			x = _et_1[_egCount_1 >> _EG_CNT_FP] + _egTLv_1 + _lfoOutAM_1;
			// pg update --->
			x = _wt_1[_pgPhase_1 >> _PG_PHASE_FP] + x;
			// OP 1 out ----->
			n += _log2vT[x];
			// counter --->
			_egCount_1 += _egCountStep_1;
			_pgPhase_1 = _pgPhase_1 + _pgPhaseStep_1 & _PG_PHASE_MASK;
		}
	// Operator 2 --->
		if (state & 0b0100)
		{
			// eg update --->
			if (_egCount_2 >= _egGate_2)
			{
				if (_shiftEGState(2))_et_2 = _egLevelTable_2;
				else
				{
					if (state & 0b1011) { state &= 0b11111011; _egCount_2 = _EG_MAX_CNT - 1; }
					else { killProcess(); return; }
				}
			}
			x = _et_2[_egCount_2 >> _EG_CNT_FP] + _egTLv_2 + _lfoOutAM_2;
			// pg update --->
			x = _wt_2[_pgPhase_2 >> _PG_PHASE_FP] + x;
			// OP 1 out ----->
			n += _log2vT[x];
			// counter --->
			_egCount_2 += _egCountStep_2;
			_pgPhase_2 = _pgPhase_2 + _pgPhaseStep_2 & _PG_PHASE_MASK;
		}
	// Operator 3 --->
		if (state & 0b1000)
		{
			// eg update --->
			if (_egCount_3 >= _egGate_3)
			{
				if (_shiftEGState(3))_et_3 = _egLevelTable_3;
				else
				{
					if (state & 0b0111) { state &= 0b11110111; _egCount_3 = _EG_MAX_CNT - 1; }
					else { killProcess(); return; }
				}
			}
			x = _et_3[_egCount_3 >> _EG_CNT_FP] + _egTLv_3 + _lfoOutAM_3;
			// pg update --->
			x = _wt_3[_pgPhase_3 >> _PG_PHASE_FP] + x;
			// OP 1 out ----->
			n += _log2vT[x];
			// counter --->
			_egCount_3 += _egCountStep_3;
			_pgPhase_3 = _pgPhase_3 + _pgPhaseStep_3 & _PG_PHASE_MASK;
		}
		// out --->
		outL[i] += n * _vOutL;
		outR[i] += n * _vOutR;
	}
}
void SynthesizerFM4::VoiceUnitFM4::_proc9(float* const& outL, float* const& outR, const int32& end)
{
	const int32* _et_0 = _egLevelTable_0, *_et_1 = _egLevelTable_1,
	            *_et_2 = _egLevelTable_2, *_et_3 = _egLevelTable_3;
	const float* const& _log2vT = _LOG2V_TABLE;
	const int32* const& _log2vTI = _LOG2V_TABLE_I;
	const int32* const& _wt_0 = _waveTable_0, * const& _wt_1 = _waveTable_1,
	           * const& _wt_2 = _waveTable_2, * const& _wt_3 = _waveTable_3;
	int x, p, f, o; float n = 0.0f;
	for (int& i = startFrame; i < end; ++i)
	{
	// lfo Update --->
		if (_lfoTimer <= 0)
		{
			p = _lfoWaveTable[_lfoPhase];
			x = (p * _lfoMod_0) >> 11;
			_lfoOutAM_0 = x & 0x7ffffffe;
			x = (p * _lfoMod_1) >> 11;
			_lfoOutAM_1 = x & 0x7ffffffe;
			x = (p * _lfoMod_2) >> 11;
			_lfoOutAM_2 = x & 0x7ffffffe;
			x = (p * _lfoMod_3) >> 11;
			_lfoOutAM_3 = x & 0x7ffffffe;
			_lfoPhase = _lfoPhase + 2 & _LFO_WAVE_TABLE_MASK;
			_lfoTimer += _LFO_TIMER_INIT_VALUE;
		}
	// Operator 0 --->
		if (state & 0b0001)
		{
			// eg update --->
			if (_egCount_0 >= _egGate_0)
			{
				if (_shiftEGState(0))_et_0 = _egLevelTable_0;
				else
				{
					if (state & 0b1110) { state &= 0b11111110; _egCount_0 = _EG_MAX_CNT - 1; }
					else { killProcess(); return; }
				}
			}
			x = _et_0[_egCount_0 >> _EG_CNT_FP] + _egTLv_0 + _lfoOutAM_0;
			// pg update --->
			p = _pgPhase_0 >> _PG_PHASE_FP;
			// FeedBack -->
			f = _log2vTI[_wt_0[p] + _fdbk];
			// OP 0 out --->
			p = _wt_0[p + f & _WAVE_TABLE_MASK] + x;
			o = _log2vTI[p];
			// counter --->	
			_egCount_0 += _egCountStep_0;
			_pgPhase_0 = _pgPhase_0 + _pgPhaseStep_0 & _PG_PHASE_MASK;
		}
		else o = 0;
	// Operator 1 --->
		if (state & 0b0010)
		{
		// eg update --->
			if (_egCount_1 >= _egGate_1)
			{
				if (_shiftEGState(1))_et_1 = _egLevelTable_1;
				else
				{
					if (state & 0b1101) { state &= 0b11111101; _egCount_1 = _EG_MAX_CNT - 1; }
					else { killProcess(); return; }
				}
			}
			x = _et_1[_egCount_1 >> _EG_CNT_FP] + _egTLv_1 + _lfoOutAM_1;
		// pg update --->
			p = (o + (_pgPhase_1 >> _PG_PHASE_FP)) & _WAVE_TABLE_MASK;
			x = _wt_1[p] + x;
		// OP 1 out ----->
			n = _log2vT[x];
		// counter --->
			_egCount_1 += _egCountStep_1;
			_pgPhase_1 = _pgPhase_1 + _pgPhaseStep_1 & _PG_PHASE_MASK;
		}
		else n = 0.0f;
	// Operator 2 --->
		if (state & 0b0100)
		{
			// eg update --->
			if (_egCount_2 >= _egGate_2)
			{
				if (_shiftEGState(2))_et_2 = _egLevelTable_2;
				else
				{
					if (state & 0b1011) { state &= 0b11111011; _egCount_2 = _EG_MAX_CNT - 1; }
					else { killProcess(); return; }
				}
			}
			x = _et_2[_egCount_2 >> _EG_CNT_FP] + _egTLv_2 + _lfoOutAM_2;
			// pg update --->
			x = _wt_2[_pgPhase_2 >> _PG_PHASE_FP] + x;
			// OP 1 out ----->
			n += _log2vT[x];
			// counter --->
			_egCount_2 += _egCountStep_2;
			_pgPhase_2 = _pgPhase_2 + _pgPhaseStep_2 & _PG_PHASE_MASK;
		}
	// Operator 3 --->
		if (state & 0b1000)
		{
			// eg update --->
			if (_egCount_3 >= _egGate_3)
			{
				if (_shiftEGState(3))_et_3 = _egLevelTable_3;
				else
				{
					if (state & 0b0111) { state &= 0b11110111; _egCount_3 = _EG_MAX_CNT - 1; }
					else { killProcess(); return; }
				}
			}
			x = _et_3[_egCount_3 >> _EG_CNT_FP] + _egTLv_3 + _lfoOutAM_3;
			// pg update --->
			x = _wt_3[_pgPhase_3 >> _PG_PHASE_FP] + x;
			// OP 1 out ----->
			n += _log2vT[x];
			// counter --->
			_egCount_3 += _egCountStep_3;
			_pgPhase_3 = _pgPhase_3 + _pgPhaseStep_3 & _PG_PHASE_MASK;
		}
	// out --->
		outL[i] += n * _vOutL;
		outR[i] += n * _vOutR;
	}
}
void SynthesizerFM4::VoiceUnitFM4::_procA(float* const& outL, float* const& outR, const int32& end)
{
	const int32* _et_0 = _egLevelTable_0, *_et_1 = _egLevelTable_1,
	            *_et_2 = _egLevelTable_2, *_et_3 = _egLevelTable_3;
	const float* const& _log2vT = _LOG2V_TABLE;
	const int32* const& _log2vTI = _LOG2V_TABLE_I;
	const int32* const& _wt_0 = _waveTable_0, * const& _wt_1 = _waveTable_1,
	           * const& _wt_2 = _waveTable_2, * const& _wt_3 = _waveTable_3;
	int x, p, f, o; float n = 0.0f;
	for (int& i = startFrame; i < end; ++i)
	{
	// lfo Update --->
		if (_lfoTimer <= 0)
		{
			p = _lfoWaveTable[_lfoPhase];
			x = (p * _lfoMod_0) >> 11;
			_lfoOutAM_0 = x & 0x7ffffffe;
			x = (p * _lfoMod_1) >> 11;
			_lfoOutAM_1 = x & 0x7ffffffe;
			x = (p * _lfoMod_2) >> 11;
			_lfoOutAM_2 = x & 0x7ffffffe;
			x = (p * _lfoMod_3) >> 11;
			_lfoOutAM_3 = x & 0x7ffffffe;
			_lfoPhase = _lfoPhase + 2 & _LFO_WAVE_TABLE_MASK;
			_lfoTimer += _LFO_TIMER_INIT_VALUE;
		}
	// Operator 0 --->
		if (state & 0b0001)
		{
			// eg update --->
			if (_egCount_0 >= _egGate_0)
			{
				if (_shiftEGState(0))_et_0 = _egLevelTable_0;
				else
				{
					if (state & 0b1110) { state &= 0b11111110; _egCount_0 = _EG_MAX_CNT - 1; }
					else { killProcess(); return; }
				}
			}
			x = _et_0[_egCount_0 >> _EG_CNT_FP] + _egTLv_0 + _lfoOutAM_0;
			// pg update --->
			p = _pgPhase_0 >> _PG_PHASE_FP;
			// FeedBack -->
			f = _log2vTI[_wt_0[p] + _fdbk];
			// OP 0 out --->
			p = _wt_0[p + f & _WAVE_TABLE_MASK] + x;
			o = _log2vTI[p];
			// counter --->	
			_egCount_0 += _egCountStep_0;
			_pgPhase_0 = _pgPhase_0 + _pgPhaseStep_0 & _PG_PHASE_MASK;
		}
		else o = 0;
	// Operator 1 --->
		if (state & 0b0010)
		{
			// eg update --->
			if (_egCount_1 >= _egGate_1)
			{
				if (_shiftEGState(1))_et_1 = _egLevelTable_1;
				else
				{
					if (state & 0b1101) { state &= 0b11111101; _egCount_1 = _EG_MAX_CNT - 1; }
					else { killProcess(); return; }
				}
			}
			x = _et_1[_egCount_1 >> _EG_CNT_FP] + _egTLv_1 + _lfoOutAM_1;
			// pg update --->
			p = (o + (_pgPhase_1 >> _PG_PHASE_FP)) & _WAVE_TABLE_MASK;
			x = _wt_1[p] + x;
			// OP 1 out ----->
			n = _log2vT[x];
			// counter --->
			_egCount_1 += _egCountStep_1;
			_pgPhase_1 = _pgPhase_1 + _pgPhaseStep_1 & _PG_PHASE_MASK;
		}
		else n = 0.0f;
	// Operator 2 --->
		if (state & 0b0100)
		{
			// eg update --->
			if (_egCount_2 >= _egGate_2)
			{
				if (_shiftEGState(2))_et_2 = _egLevelTable_2;
				else
				{
					if (state & 0b1011) { state &= 0b11111011; _egCount_2 = _EG_MAX_CNT - 1; }
					else { killProcess(); return; }
				}
			}
			x = _et_2[_egCount_2 >> _EG_CNT_FP] + _egTLv_2 + _lfoOutAM_2;
			// pg update --->
			p = (o + (_pgPhase_2 >> _PG_PHASE_FP)) & _WAVE_TABLE_MASK;
			x = _wt_2[p] + x;
			// OP 1 out ----->
			n += _log2vT[x];
			// counter --->
			_egCount_2 += _egCountStep_2;
			_pgPhase_2 = _pgPhase_2 + _pgPhaseStep_2 & _PG_PHASE_MASK;
		}
	// Operator 3 --->
		if (state & 0b1000)
		{
			// eg update --->
			if (_egCount_3 >= _egGate_3)
			{
				if (_shiftEGState(3))_et_3 = _egLevelTable_3;
				else
				{
					if (state & 0b0111) { state &= 0b11110111; _egCount_3 = _EG_MAX_CNT - 1; }
					else { killProcess(); return; }
				}
			}
			x = _et_3[_egCount_3 >> _EG_CNT_FP] + _egTLv_3 + _lfoOutAM_3;
			// pg update --->
			p = (o + (_pgPhase_3 >> _PG_PHASE_FP)) & _WAVE_TABLE_MASK;
			x = _wt_3[p] + x;
			// OP 1 out ----->
			n += _log2vT[x];
			// counter --->
			_egCount_3 += _egCountStep_3;
			_pgPhase_3 = _pgPhase_3 + _pgPhaseStep_3 & _PG_PHASE_MASK;
		}
	// out --->
		outL[i] += n * _vOutL;
		outR[i] += n * _vOutR;
	// counter --->
		_lfoTimer -= _lfoTimerStep;
	}
}
void SynthesizerFM4::VoiceUnitFM4::_procB(float* const& outL, float* const& outR, const int32& end)
{
	const int32* _et_0 = _egLevelTable_0, *_et_1 = _egLevelTable_1,
	            *_et_2 = _egLevelTable_2, *_et_3 = _egLevelTable_3;
	const float* const& _log2vT = _LOG2V_TABLE;
	const int32* const& _log2vTI = _LOG2V_TABLE_I;
	const int32* const& _wt_0 = _waveTable_0, * const& _wt_1 = _waveTable_1,
	           * const& _wt_2 = _waveTable_2, * const& _wt_3 = _waveTable_3;
	int x, p, f, o; float n = 0.0f;
	for (int& i = startFrame; i < end; ++i)
	{
	// lfo Update --->
		if (_lfoTimer <= 0)
		{
			p = _lfoWaveTable[_lfoPhase];
			x = (p * _lfoMod_0) >> 11;
			_lfoOutAM_0 = x & 0x7ffffffe;
			x = (p * _lfoMod_1) >> 11;
			_lfoOutAM_1 = x & 0x7ffffffe;
			x = (p * _lfoMod_2) >> 11;
			_lfoOutAM_2 = x & 0x7ffffffe;
			x = (p * _lfoMod_3) >> 11;
			_lfoOutAM_3 = x & 0x7ffffffe;
			_lfoPhase = _lfoPhase + 2 & _LFO_WAVE_TABLE_MASK;
			_lfoTimer += _LFO_TIMER_INIT_VALUE;
		}
	// Operator 0 --->
		if (state & 0b0001)
		{
			// eg update --->
			if (_egCount_0 >= _egGate_0)
			{
				if (_shiftEGState(0))_et_0 = _egLevelTable_0;
				else
				{
					if (state & 0b1110) { state &= 0b11111110; _egCount_0 = _EG_MAX_CNT - 1; }
					else { killProcess(); return; }
				}
			}
			x = _et_0[_egCount_0 >> _EG_CNT_FP] + _egTLv_0 + _lfoOutAM_0;
			// pg update --->
			p = _pgPhase_0 >> _PG_PHASE_FP;
			// FeedBack -->
			f = _log2vTI[_wt_0[p] + _fdbk];
			// OP 0 out --->
			x = _wt_0[p + f & _WAVE_TABLE_MASK] + x;
			o = _log2vTI[x];
			// counter --->	
			_egCount_0 += _egCountStep_0;
			_pgPhase_0 = _pgPhase_0 + _pgPhaseStep_0 & _PG_PHASE_MASK;
		}
		else o = 0;
	// Operator 1 --->
		if (state & 0b0010)
		{
			// eg update --->
			if (_egCount_1 >= _egGate_1)
			{
				if (_shiftEGState(1))_et_1 = _egLevelTable_1;
				else
				{
					if (state & 0b1101) { state &= 0b11111101; _egCount_1 = _EG_MAX_CNT - 1; }
					else { killProcess(); return; }
				}
			}
			x = _et_1[_egCount_1 >> _EG_CNT_FP] + _egTLv_1 + _lfoOutAM_1;
			// pg update --->
			p = (o + (_pgPhase_1 >> _PG_PHASE_FP)) & _WAVE_TABLE_MASK;
			x = _wt_1[p] + x;
			// OP 1 out ----->
			n = _log2vT[x];
			// counter --->
			_egCount_1 += _egCountStep_1;
			_pgPhase_1 = _pgPhase_1 + _pgPhaseStep_1 & _PG_PHASE_MASK;
		}
		else n = 0.0f;
	// Operator 2 --->
		if (state & 0b0100)
		{
			// eg update --->
			if (_egCount_2 >= _egGate_2)
			{
				if (_shiftEGState(2))_et_2 = _egLevelTable_2;
				else
				{
					if (state & 0b1011) { state &= 0b11111011; _egCount_2 = _EG_MAX_CNT - 1; }
					else { killProcess(); return; }
				}
			}
			x = _et_2[_egCount_2 >> _EG_CNT_FP] + _egTLv_2 + _lfoOutAM_2;
			// pg update --->
			x = _wt_2[_pgPhase_2 >> _PG_PHASE_FP] + x;
			// OP 0 out --->
			o = _log2vTI[x];
			// counter --->	
			_egCount_2 += _egCountStep_2;
			_pgPhase_2 = _pgPhase_2 + _pgPhaseStep_2 & _PG_PHASE_MASK;
		}
		else o = 0;
	// Operator 3 --->
		if (state & 0b1000)
		{
			// eg update --->
			if (_egCount_3 >= _egGate_3)
			{
				if (_shiftEGState(3))_et_3 = _egLevelTable_3;
				else
				{
					if (state & 0b0111) { state &= 0b11110111; _egCount_3 = _EG_MAX_CNT - 1; }
					else { killProcess(); return; }
				}
			}
			x = _et_3[_egCount_3 >> _EG_CNT_FP] + _egTLv_3 + _lfoOutAM_3;
			// pg update --->
			p = (o + (_pgPhase_3 >> _PG_PHASE_FP)) & _WAVE_TABLE_MASK;
			x = _wt_3[p] + x;
			// OP 1 out ----->
			n += _log2vT[x];
			// counter --->
			_egCount_3 += _egCountStep_3;
			_pgPhase_3 = _pgPhase_3 + _pgPhaseStep_3 & _PG_PHASE_MASK;
		}
		// out --->
		outL[i] += n * _vOutL;
		outR[i] += n * _vOutR;
	// counter --->
		_lfoTimer -= _lfoTimerStep;
	}
}
void SynthesizerFM4::VoiceUnitFM4::_procC(float* const& outL, float* const& outR, const int32& end)
{
	const int32* _et_0 = _egLevelTable_0, *_et_1 = _egLevelTable_1,
	            *_et_2 = _egLevelTable_2, *_et_3 = _egLevelTable_3;
	const float* const& _log2vT = _LOG2V_TABLE;
	const int32* const& _log2vTI = _LOG2V_TABLE_I;
	const int32* const& _wt_0 = _waveTable_0, * const& _wt_1 = _waveTable_1,
	           * const& _wt_2 = _waveTable_2, * const& _wt_3 = _waveTable_3;
	int x, p, f, o; float n = 0.0f;
	for (int& i = startFrame; i < end; ++i)
	{
	// lfo Update --->
		if (_lfoTimer <= 0)
		{
			p = _lfoWaveTable[_lfoPhase];
			x = (p * _lfoMod_0) >> 11;
			_lfoOutAM_0 = x & 0x7ffffffe;
			x = (p * _lfoMod_1) >> 11;
			_lfoOutAM_1 = x & 0x7ffffffe;
			x = (p * _lfoMod_2) >> 11;
			_lfoOutAM_2 = x & 0x7ffffffe;
			x = (p * _lfoMod_3) >> 11;
			_lfoOutAM_3 = x & 0x7ffffffe;
			_lfoPhase = _lfoPhase + 2 & _LFO_WAVE_TABLE_MASK;
			_lfoTimer += _LFO_TIMER_INIT_VALUE;
		}
	// Operator 0 --->
		if (state & 0b0001)
		{
			// eg update --->
			if (_egCount_0 >= _egGate_0)
			{
				if (_shiftEGState(0))_et_0 = _egLevelTable_0;
				else
				{
					if (state & 0b1110) { state &= 0b11111110; _egCount_0 = _EG_MAX_CNT - 1; }
					else { killProcess(); return; }
				}
			}
			x = _et_0[_egCount_0 >> _EG_CNT_FP] + _egTLv_0 + _lfoOutAM_0;
			// pg update --->
			p = _pgPhase_0 >> _PG_PHASE_FP;
			// FeedBack -->
			f = _log2vTI[_wt_0[p] + _fdbk];
			// OP 0 out --->
			x = _wt_0[p + f & _WAVE_TABLE_MASK] + x;
			o = _log2vTI[x];
			// counter --->	
			_egCount_0 += _egCountStep_0;
			_pgPhase_0 = _pgPhase_0 + _pgPhaseStep_0 & _PG_PHASE_MASK;
		}
		else o = 0;
	// Operator 1 --->
		if (state & 0b0010)
		{
		// eg update --->
			if (_egCount_1 >= _egGate_1)
			{
				if (_shiftEGState(1))_et_1 = _egLevelTable_1;
				else
				{
					if (state & 0b1101) { state &= 0b11111101; _egCount_1 = _EG_MAX_CNT - 1; }
					else { killProcess(); return; }
				}
			}
			x = _et_1[_egCount_1 >> _EG_CNT_FP] + _egTLv_1 + _lfoOutAM_1;
		// pg update --->
			p = (o + (_pgPhase_1 >> _PG_PHASE_FP)) & _WAVE_TABLE_MASK;
			x = _wt_1[p] + x;
		// OP 1 out ----->
			o = _log2vTI[x];
		// counter --->
			_egCount_1 += _egCountStep_1;
			_pgPhase_1 = _pgPhase_1 + _pgPhaseStep_1 & _PG_PHASE_MASK;
		}
		else o = 0;
	// Operator 2 --->
		if (state & 0b0100)
		{
			// eg update --->
			if (_egCount_2 >= _egGate_2)
			{
				if (_shiftEGState(2))_et_2 = _egLevelTable_2;
				else
				{
					if (state & 0b1011) { state &= 0b11111011; _egCount_2 = _EG_MAX_CNT - 1; }
					else { killProcess(); return; }
				}
			}
			x = _et_2[_egCount_2 >> _EG_CNT_FP] + _egTLv_2 + _lfoOutAM_2;
			// pg update --->
			x = _wt_2[_pgPhase_2 >> _PG_PHASE_FP] + x;
			// OP 1 out ----->
			o += _log2vTI[x];
			// counter --->
			_egCount_2 += _egCountStep_2;
			_pgPhase_2 = _pgPhase_2 + _pgPhaseStep_2 & _PG_PHASE_MASK;
		}
	// Operator 3 --->
		if (state & 0b1000)
		{
			// eg update --->
			if (_egCount_3 >= _egGate_3)
			{
				if (_shiftEGState(3))_et_3 = _egLevelTable_3;
				else { killProcess(); return; }
			}
			x = _et_3[_egCount_3 >> _EG_CNT_FP] + _egTLv_3 + _lfoOutAM_3;
			// pg update --->
			p = (o + (_pgPhase_3 >> _PG_PHASE_FP)) & _WAVE_TABLE_MASK;
			x = _wt_3[p] + x;
			// OP 1 out ----->
			n = _log2vT[x];
			// counter --->
			_egCount_3 += _egCountStep_3;
			_pgPhase_3 = _pgPhase_3 + _pgPhaseStep_3 & _PG_PHASE_MASK;
		}
	// out --->
		outL[i] += n * _vOutL;
		outR[i] += n * _vOutR;
	}
}
void SynthesizerFM4::VoiceUnitFM4::_procD(float* const& outL, float* const& outR, const int32& end)
{
	const int32* _et_0 = _egLevelTable_0, *_et_1 = _egLevelTable_1,
	            *_et_2 = _egLevelTable_2, *_et_3 = _egLevelTable_3;
	const float* const& _log2vT = _LOG2V_TABLE;
	const int32* const& _log2vTI = _LOG2V_TABLE_I;
	const int32* const& _wt_0 = _waveTable_0, * const& _wt_1 = _waveTable_1,
	           * const& _wt_2 = _waveTable_2, * const& _wt_3 = _waveTable_3;
	int x, p, f, o; float n = 0.0f;
	for (int& i = startFrame; i < end; ++i)
	{
	// lfo Update --->
		if (_lfoTimer <= 0)
		{
			p = _lfoWaveTable[_lfoPhase];
			x = (p * _lfoMod_0) >> 11;
			_lfoOutAM_0 = x & 0x7ffffffe;
			x = (p * _lfoMod_1) >> 11;
			_lfoOutAM_1 = x & 0x7ffffffe;
			x = (p * _lfoMod_2) >> 11;
			_lfoOutAM_2 = x & 0x7ffffffe;
			x = (p * _lfoMod_3) >> 11;
			_lfoOutAM_3 = x & 0x7ffffffe;
			_lfoPhase = _lfoPhase + 2 & _LFO_WAVE_TABLE_MASK;
			_lfoTimer += _LFO_TIMER_INIT_VALUE;
		}
	// Operator 1 --->
		if (state & 0b0010)
		{
			// eg update --->
			if (_egCount_1 >= _egGate_1)
			{
				if (_shiftEGState(1))_et_1 = _egLevelTable_1;
				else
				{
					if (state & 0b1101) { state &= 0b11111101; _egCount_1 = _EG_MAX_CNT - 1; }
					else { killProcess(); return; }
				}
			}
			x = _et_1[_egCount_1 >> _EG_CNT_FP] + _egTLv_1 + _lfoOutAM_1;
			// pg update --->
			p = _wt_1[_pgPhase_1 >> _PG_PHASE_FP] + x;
			// OP 1 out ----->
			o = _log2vTI[p];
			// counter --->
			_egCount_1 += _egCountStep_1;
			_pgPhase_1 = _pgPhase_1 + _pgPhaseStep_1 & _PG_PHASE_MASK;
		}
		else o = 0;
	// Operator 2 --->
		if (state & 0b0100)
		{
			// eg update --->
			if (_egCount_2 >= _egGate_2)
			{
				if (_shiftEGState(2))_et_2 = _egLevelTable_2;
				else
				{
					if (state & 0b1011) { state &= 0b11111011; _egCount_2 = _EG_MAX_CNT - 1; }
					else { killProcess(); return; }
				}
			}
			x = _et_2[_egCount_2 >> _EG_CNT_FP] + _egTLv_2 + _lfoOutAM_2;
			// pg update --->
			p = (o + (_pgPhase_2 >> _PG_PHASE_FP)) & _WAVE_TABLE_MASK;
			x = _wt_2[p] + x;
			// OP 1 out ----->
			o = _log2vTI[x];
			// counter --->
			_egCount_2 += _egCountStep_2;
			_pgPhase_2 = _pgPhase_2 + _pgPhaseStep_2 & _PG_PHASE_MASK;
		}
		else o = 0;
	// Operator 0 --->
		if (state & 0b0001)
		{
			// eg update --->
			if (_egCount_0 >= _egGate_0)
			{
				if (_shiftEGState(0))_et_0 = _egLevelTable_0;
				else
				{
					if (state & 0b1110) { state &= 0b11111110; _egCount_0 = _EG_MAX_CNT - 1; }
					else { killProcess(); return; }
				}
			}
			x = _et_0[_egCount_0 >> _EG_CNT_FP] + _egTLv_0 + _lfoOutAM_0;
			// pg update --->
			p = _pgPhase_0 >> _PG_PHASE_FP;
			// FeedBack -->
			f = _log2vTI[_wt_0[p] + _fdbk];
			// OP 0 out --->
			x = _wt_0[p + f & _WAVE_TABLE_MASK] + x;
			o += _log2vTI[x];
			// counter --->	
			_egCount_0 += _egCountStep_0;
			_pgPhase_0 = _pgPhase_0 + _pgPhaseStep_0 & _PG_PHASE_MASK;
		}
	// Operator 3 --->
		if (state & 0b1000)
		{
			// eg update --->
			if (_egCount_3 >= _egGate_3)
			{
				if (_shiftEGState(3))_et_3 = _egLevelTable_3;
				else { killProcess(); return; }
			}
			x = _et_3[_egCount_3 >> _EG_CNT_FP] + _egTLv_3 + _lfoOutAM_3;
			// pg update --->
			p = (o + (_pgPhase_3 >> _PG_PHASE_FP)) & _WAVE_TABLE_MASK;
			x = _wt_3[p] + x;
			// OP 1 out ----->
			n = _log2vT[x];
			// counter --->
			_egCount_3 += _egCountStep_3;
			_pgPhase_3 = _pgPhase_3 + _pgPhaseStep_3 & _PG_PHASE_MASK;
		}
		// out --->
		outL[i] += n * _vOutL;
		outR[i] += n * _vOutR;
	}
}
void SynthesizerFM4::VoiceUnitFM4::_procE(float* const& outL, float* const& outR, const int32& end)
{
	const int32* _et_0 = _egLevelTable_0, *_et_1 = _egLevelTable_1,
	            *_et_2 = _egLevelTable_2, *_et_3 = _egLevelTable_3;
	const float* const& _log2vT = _LOG2V_TABLE;
	const int32* const& _log2vTI = _LOG2V_TABLE_I;
	const int32* const& _wt_0 = _waveTable_0, * const& _wt_1 = _waveTable_1,
	           * const& _wt_2 = _waveTable_2, * const& _wt_3 = _waveTable_3;
	int x, p, f, o; float n = 0.0f;
	for (int& i = startFrame; i < end; ++i)
	{
	// lfo Update --->
		if (_lfoTimer <= 0)
		{
			p = _lfoWaveTable[_lfoPhase];
			x = (p * _lfoMod_0) >> 11;
			_lfoOutAM_0 = x & 0x7ffffffe;
			x = (p * _lfoMod_1) >> 11;
			_lfoOutAM_1 = x & 0x7ffffffe;
			x = (p * _lfoMod_2) >> 11;
			_lfoOutAM_2 = x & 0x7ffffffe;
			x = (p * _lfoMod_3) >> 11;
			_lfoOutAM_3 = x & 0x7ffffffe;
			_lfoPhase = _lfoPhase + 2 & _LFO_WAVE_TABLE_MASK;
			_lfoTimer += _LFO_TIMER_INIT_VALUE;
		}
	// Operator 0 --->
		if (state & 0b0001)
		{
			// eg update --->
			if (_egCount_0 >= _egGate_0)
			{
				if (_shiftEGState(0))_et_0 = _egLevelTable_0;
				else
				{
					if (state & 0b1110) { state &= 0b11111110; _egCount_0 = _EG_MAX_CNT - 1; }
					else { killProcess(); return; }
				}
			}
			x = _et_0[_egCount_0 >> _EG_CNT_FP] + _egTLv_0 + _lfoOutAM_0;
			// pg update --->
			p = _pgPhase_0 >> _PG_PHASE_FP;
			// FeedBack -->
			f = _log2vTI[_wt_0[p] + _fdbk];
			// OP 0 out --->
			x = _wt_0[p + f & _WAVE_TABLE_MASK] + x;
			o = _log2vTI[x];
			// counter --->	
			_egCount_0 += _egCountStep_0;
			_pgPhase_0 = _pgPhase_0 + _pgPhaseStep_0 & _PG_PHASE_MASK;
		}
		else o = 0;
	// Operator 1 --->
		if (state & 0b0010)
		{
		// eg update --->
			if (_egCount_1 >= _egGate_1)
			{
				if (_shiftEGState(1))_et_1 = _egLevelTable_1;
				else
				{
					if (state & 0b1101) { state &= 0b11111101; _egCount_1 = _EG_MAX_CNT - 1; }
					else { killProcess(); return; }
				}
			}
			x = _et_1[_egCount_1 >> _EG_CNT_FP] + _egTLv_1 + _lfoOutAM_1;
		// pg update --->
			x = _wt_1[_pgPhase_1 >> _PG_PHASE_FP] + x;
		// OP 1 out ----->
			o += _log2vTI[x];
		// counter --->
			_egCount_1 += _egCountStep_1;
			_pgPhase_1 = _pgPhase_1 + _pgPhaseStep_1 & _PG_PHASE_MASK;
		}
	// Operator 2 --->
		if (state & 0b0100)
		{
			// eg update --->
			if (_egCount_2 >= _egGate_2)
			{
				if (_shiftEGState(2))_et_2 = _egLevelTable_2;
				else
				{
					if (state & 0b1011) { state &= 0b11111011; _egCount_2 = _EG_MAX_CNT - 1; }
					else { killProcess(); return; }
				}
			}
			x = _et_2[_egCount_2 >> _EG_CNT_FP] + _egTLv_2 + _lfoOutAM_2;
			// pg update --->
			p = (o + (_pgPhase_2 >> _PG_PHASE_FP)) & _WAVE_TABLE_MASK;
			x = _wt_2[p] + x;
			// OP 1 out ----->
			o = _log2vTI[x];
			// counter --->
			_egCount_2 += _egCountStep_2;
			_pgPhase_2 = _pgPhase_2 + _pgPhaseStep_2 & _PG_PHASE_MASK;
		}
		else o = 0;
	// Operator 3 --->
		if (state & 0b1000)
		{
			// eg update --->
			if (_egCount_3 >= _egGate_3)
			{
				if (_shiftEGState(3))_et_3 = _egLevelTable_3;
				else { killProcess(); return; }
			}
			x = _et_3[_egCount_3 >> _EG_CNT_FP] + _egTLv_3 + _lfoOutAM_3;
			// pg update --->
			p = (o + (_pgPhase_3 >> _PG_PHASE_FP)) & _WAVE_TABLE_MASK;
			x = _wt_3[p] + x;
			// OP 1 out ----->
			n = _log2vT[x];
			// counter --->
			_egCount_3 += _egCountStep_3;
			_pgPhase_3 = _pgPhase_3 + _pgPhaseStep_3 & _PG_PHASE_MASK;
		}
	// out --->
		outL[i] += n * _vOutL;
		outR[i] += n * _vOutR;
	// counter --->
		_lfoTimer -= _lfoTimerStep;
	}
}
void SynthesizerFM4::VoiceUnitFM4::_procF(float* const& outL, float* const& outR, const int32& end)
{
	const int32* _et_0 = _egLevelTable_0, *_et_1 = _egLevelTable_1,
	            *_et_2 = _egLevelTable_2, *_et_3 = _egLevelTable_3;
	const float* const& _log2vT = _LOG2V_TABLE;
	const int32* const& _log2vTI = _LOG2V_TABLE_I;
	const int32* const& _wt_0 = _waveTable_0, * const& _wt_1 = _waveTable_1,
	           * const& _wt_2 = _waveTable_2, * const& _wt_3 = _waveTable_3;
	int x, p, f, o; float n = 0.0f;
	for (int& i = startFrame; i < end; ++i)
	{
	// lfo Update --->
		if (_lfoTimer <= 0)
		{
			p = _lfoWaveTable[_lfoPhase];
			x = (p * _lfoMod_0) >> 11;
			_lfoOutAM_0 = x & 0x7ffffffe;
			x = (p * _lfoMod_1) >> 11;
			_lfoOutAM_1 = x & 0x7ffffffe;
			x = (p * _lfoMod_2) >> 11;
			_lfoOutAM_2 = x & 0x7ffffffe;
			x = (p * _lfoMod_3) >> 11;
			_lfoOutAM_3 = x & 0x7ffffffe;
			_lfoPhase = _lfoPhase + 2 & _LFO_WAVE_TABLE_MASK;
			_lfoTimer += _LFO_TIMER_INIT_VALUE;
		}
	// Operator 0 --->
		if (state & 0b0001)
		{
			// eg update --->
			if (_egCount_0 >= _egGate_0)
			{
				if (_shiftEGState(0))_et_0 = _egLevelTable_0;
				else
				{
					if (state & 0b1110) { state &= 0b11111110; _egCount_0 = _EG_MAX_CNT - 1; }
					else { killProcess(); return; }
				}
			}
			x = _et_0[_egCount_0 >> _EG_CNT_FP] + _egTLv_0 + _lfoOutAM_0;
			// pg update --->
			p = _pgPhase_0 >> _PG_PHASE_FP;
			// FeedBack -->
			f = _log2vTI[_wt_0[p] + _fdbk];
			// OP 0 out --->
			p = _wt_0[p + f & _WAVE_TABLE_MASK] + x;
			o = _log2vTI[p];
			// counter --->	
			_egCount_0 += _egCountStep_0;
			_pgPhase_0 = _pgPhase_0 + _pgPhaseStep_0 & _PG_PHASE_MASK;
		}
		else o = 0;
	// Operator 1 --->
		if (state & 0b0010)
		{
		// eg update --->
			if (_egCount_1 >= _egGate_1)
			{
				if (_shiftEGState(1))_et_1 = _egLevelTable_1;
				else
				{
					if (state & 0b1101) { state &= 0b11111101; _egCount_1 = _EG_MAX_CNT - 1; }
					else { killProcess(); return; }
				}
			}
			x = _et_1[_egCount_1 >> _EG_CNT_FP] + _egTLv_1 + _lfoOutAM_1;
		// pg update --->
			p = (o + (_pgPhase_1 >> _PG_PHASE_FP)) & _WAVE_TABLE_MASK;
			p = _wt_1[p] + x;
		// OP 1 out ----->
			o = _log2vTI[p];
		// counter --->
			_egCount_1 += _egCountStep_1;
			_pgPhase_1 = _pgPhase_1 + _pgPhaseStep_1 & _PG_PHASE_MASK;
		}
		else o = 0;
	// Operator 2 --->
		if (state & 0b0100)
		{
			// eg update --->
			if (_egCount_2 >= _egGate_2)
			{
				if (_shiftEGState(2))_et_2 = _egLevelTable_2;
				else
				{
					if (state & 0b1011) { state &= 0b11111011; _egCount_2 = _EG_MAX_CNT - 1; }
					else { killProcess(); return; }
				}
			}
			x = _et_2[_egCount_2 >> _EG_CNT_FP] + _egTLv_2 + _lfoOutAM_2;
			// pg update --->
			p = (o + (_pgPhase_2 >> _PG_PHASE_FP)) & _WAVE_TABLE_MASK;
			p = _wt_2[p] + x;
			// OP 1 out ----->
			o = _log2vTI[p];
			// counter --->
			_egCount_2 += _egCountStep_2;
			_pgPhase_2 = _pgPhase_2 + _pgPhaseStep_2 & _PG_PHASE_MASK;
		}
		else o = 0;
	// Operator 3 --->
		if (state & 0b1000)
		{
			// eg update --->
			if (_egCount_3 >= _egGate_3)
			{
				if (_shiftEGState(3))_et_3 = _egLevelTable_3;
				else { killProcess(); return; }
			}
			x = _et_3[_egCount_3 >> _EG_CNT_FP] + _egTLv_3 + _lfoOutAM_3;
			// pg update --->
			p = (o + (_pgPhase_3 >> _PG_PHASE_FP)) & _WAVE_TABLE_MASK;
			p = _wt_3[p] + x;
			// OP 1 out ----->
			n = _log2vT[p];
			// counter --->
			_egCount_3 += _egCountStep_3;
			_pgPhase_3 = _pgPhase_3 + _pgPhaseStep_3 & _PG_PHASE_MASK;
		}
		//else n = 0.0f;
	// out --->
		outL[i] += n * _vOutL;
		outR[i] += n * _vOutR;
	// counter --->
		_lfoTimer -= _lfoTimerStep;
	}
}

void SynthesizerFM4::VoiceUnitFM4::_procSmpl1ch(float* const& outL, float* const& outR, const int32& end)
{
	const int32* _et = _egLevelTable_0;
	const float* const& _log2vT = _LOG2V_TABLE;
	const int32* const& _wt = _waveTable_0;
	for (int& i = startFrame; i < end; ++i)
	{
	// eg update --->
		if (_egCount_0 >= _egGate_0)
		{
			if (_shiftEGState(0))_et = _egLevelTable_0;
			else { killProcess(); return; }
		}
		int v = _et[_egCount_0 >> _EG_CNT_FP] + _egTLv_0;	
	// pg update --->
		if (_pgPhase_0 >= _loopE)
		{
			if (_loopF > 0) do{ _pgPhase_0 -= _loopF; } while (_pgPhase_0 >= _loopE);
			else { killProcess(); return; }
		}
		v = _wt[_pgPhase_0] + v;
		float n = _log2vT[v];
	// out --->
		outL[i] += n * _vOutL;
		outR[i] += n * _vOutR;
	// counter --->
		_egCount_0 += _egCountStep_0;
		_pgPhaseAFP += _pgPhaseStep_0;
		_pgPhase_0 += _pgPhaseAFP >> _PG_PHASE_FP;
		_pgPhaseAFP &= _PG_PHASE_FP_MASK;
	}
}
void SynthesizerFM4::VoiceUnitFM4::_procSmpl2ch(float* const& outL, float* const& outR, const int32& end)
{
	const int32* _et = _egLevelTable_0;
	const float* const& _log2vT = _LOG2V_TABLE;
	const int32* const& _wt = _waveTable_0;
	for (int& i = startFrame; i < end; ++i)
	{
	// eg update --->
		if (_egCount_0 >= _egGate_0)
		{
			if (_shiftEGState(0))_et = _egLevelTable_0;
			else { killProcess(); return; }
		}
		int v = _et[_egCount_0 >> _EG_CNT_FP] + _egTLv_0;	
	// pg update --->
		if (_pgPhase_0 >= _loopE)
		{
			if (_loopF > 0) do{ _pgPhase_0 -= _loopF; } while (_pgPhase_0 >= _loopE);
			else { killProcess(); return; }
		}
	// out --->
		int p = _pgPhase_0 << 1;
		outL[i] += _log2vT[_wt[p    ] + v] * _vOutL;
		outR[i] += _log2vT[_wt[p + 1] + v] * _vOutR;
	// counter --->
		_egCount_0 += _egCountStep_0;
		_pgPhaseAFP += _pgPhaseStep_0;
		_pgPhase_0 += _pgPhaseAFP >> _PG_PHASE_FP;
		_pgPhaseAFP &= _PG_PHASE_FP_MASK;
	}
}
void SynthesizerFM4::VoiceUnitFM4::setup(SynthData* synthData)
{
	// volume Table
	_mainVolTable = synthData->getVOLUME_TABLE_0dB();
	_exprVolTable = synthData->getVOLUME_TABLE_0dB();
	_veloVolTable = synthData->getVOLUME_TABLE_0dB();
	_panTable = synthData->getPAN_TABLE_SIN();
	// Log2 Volume Table
	_LOG2V_TABLE = synthData->getLOG2V_TABLE();
	_LOG2V_TABLE_I = synthData->getLOG2V_TABLE_I();
	// Pow2 Table
	_POW2_TABLE = synthData->getPOW2_TABLE();
	_POW2T_RS_TABLE = synthData->getPOW2T_RS_TABLE();
	_POW2T_KEY_TABLE = synthData->getPOW2T_KEY_TABLE();
	// EG
	_EG_TOTAL_LEVEL_TABLE = synthData->getEG_TOTAL_LEVEL_TABLE();
	_EG_LEVEL_TABLE_UP = synthData->getEG_LEVEL_TABLE_UP();
	_EG_LEVEL_TABLE_DW = synthData->getEG_LEVEL_TABLE_DW();
	_EG_KEY_TABLE = synthData->getEG_KEY_TABLE();
	_EG_POW2T_RS_TABLE = synthData->getEG_POW2T_RS_TABLE();
	_EG_CNT_GATE_TABLE = synthData->getEG_CNT_GATE_TABLE();
	// etc
	_MUL_TABLE = synthData->getMUL_TABLE();
	_LFO_WAVE_TABLES = synthData->getLFO_WAVE_TABLES();
}

}}}
